#include "ConnectionHandle.h"
#include "ConnectionServiceConfiguration.h"
#include "CoralCommon/IDevConnection.h"
#include "CoralCommon/Utilities.h"
#include "RelationalAccess/IConnection.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/SessionException.h"
#include "SealKernel/MessageStream.h"
#include "uuid/uuid.h"


/// default constructor
coral::ConnectionService::ConnectionHandle::ConnectionHandle():
  m_connection(),
  m_info( new ConnectionSharedInfo() ){
}

/// constructor
coral::ConnectionService::ConnectionHandle::ConnectionHandle( IConnection* connection,
                                                              const ConnectionParams& connectionParams,
                                                              const ConnectionServiceConfiguration& configuration):
  m_connection( connection ),
  m_info( new ConnectionSharedInfo(connectionParams,configuration) ){
  uuid_t guid;
  ::uuid_generate_time( guid );
  char strGuid[37];
  ::uuid_unparse( guid,strGuid );
  m_info->m_connectionId = std::string(strGuid);
}

/// destructor
coral::ConnectionService::ConnectionHandle::~ConnectionHandle(){
}

/// copy constructor
coral::ConnectionService::ConnectionHandle::ConnectionHandle( const ConnectionHandle& rhs ):
  m_connection(rhs.m_connection),
  m_info(rhs.m_info){
}

/// assignment operator
coral::ConnectionService::ConnectionHandle&
coral::ConnectionService::ConnectionHandle::operator=( const ConnectionHandle& rhs ){
  if(this!=&rhs){
    m_connection = rhs.m_connection;
    m_info = rhs.m_info;
  }
  return *this;
}

/// returns 
coral::ConnectionService::ConnectionHandle::operator bool() const{
  return m_connection.get()!=0;
}

bool coral::ConnectionService::ConnectionHandle::open(){
  if(m_connection){
     seal::MessageStream log( m_info->m_configuration.serviceContext(), m_info->m_configuration.serviceName(), seal::Msg::Verbose );
     if( ! this->isOpen() ) {
       m_info->m_open = false;
       seal::Time startConnectionAttempt = seal::Time::current();
       log << seal::Msg::Debug << "Connection parameters are: Retry-Period="<<m_info->m_configuration.connectionRetrialPeriod()<<"sec, Retry-Timeout="<<m_info->m_configuration.connectionRetrialTimeOut()<<"sec, Connection-Timeout="<<m_info->m_configuration.connectionTimeOut()<<"sec."<<seal::flush;
       // retrial block (only for ServerExceptions!)
       while(!m_info->m_open){
         try {
           m_connection->connect();
           m_info->m_open = true;
           log << seal::Msg::Info << " Connection to service \"" << m_info->m_serviceName << "\" established. Id="<< m_info->m_connectionId <<seal::flush;
          } catch (const coral::ConnectionException& exc){
            log << seal::Msg::Warning << "Failed to connect to service " << m_info->m_serviceName << ": "<< exc.what() << seal::flush;
          }
          if(!m_info->m_open){
            if ( (seal::Time::current()-startConnectionAttempt).seconds() < m_info->m_configuration.connectionRetrialTimeOut() ){
              coral::sleepSeconds(m_info->m_configuration.connectionRetrialPeriod());
            } else {
              break;
            }
          }
       }
     }
  }
  return m_info->m_open;
}

/// finalize (and invalidate) the connection.
bool coral::ConnectionService::ConnectionHandle::close(){
  bool ret = false;
  if(m_connection){
    m_connection->disconnect();
  }
  return ret;
}

/// returns true if the connection is valid
bool coral::ConnectionService::ConnectionHandle::isOpen() const {
  if(m_connection==0){
    return false;
  } else {
    return m_info->m_open;
  }
}

/// returns true if the connection is valid
bool coral::ConnectionService::ConnectionHandle::isValid() const {
  if(m_connection==0){
    return false;
  } else {
    return m_connection->isConnected(true);
  }
}

/// initialize the session
coral::ISession*  coral::ConnectionService::ConnectionHandle::newSession( const std::string& schemaName,
                                                             const std::string& userName,
                                                             const std::string& password,
                                                             coral::AccessMode accessMode){
  coral::ISession* physicalSession = 0;
  bool started = false;
  if( isOpen() ){
    seal::MessageStream log( m_info->m_configuration.serviceContext(), m_info->m_configuration.serviceName(), seal::Msg::Verbose );
    seal::Time startSessionAttempt = seal::Time::current();
    physicalSession = m_connection->newSession( schemaName, accessMode);
    if(physicalSession){
      try{
        // retrial block (only for SessionExceptions!)
        while(!started){
          try {
            physicalSession->startUserSession( userName, password );
            if(m_info->m_configuration.isMonitoringEnabled()) {
              physicalSession->monitoring().start( m_info->m_configuration.monitoringLevel() );
            }
            m_info->m_idle = false;
            started = true;
            log << seal::Msg::Info << "New session on connection to service \"" << m_info->m_serviceName << "\" started for user \"" <<userName<<"\". Connection Id="<< m_info->m_connectionId << seal::flush;
           } catch (const coral::StartSessionException& exc){
             log << seal::Msg::Warning << "Failure while attempting to start a session on connection to service " << m_info->m_serviceName << ": "<< exc.what() << seal::flush;
           }
           if( !started ){
              if ( (seal::Time::current()-startSessionAttempt).seconds() < m_info->m_configuration.connectionRetrialTimeOut() ){
                coral::sleepSeconds(m_info->m_configuration.connectionRetrialPeriod());
              } else {
                delete physicalSession;
                physicalSession = 0;
                break;
              }
           }
        }
      } catch (const coral::AuthenticationFailureException& exc){
        delete physicalSession;
        physicalSession = 0;
        throw;
      }
    }
  }
  return physicalSession;
}

///returns the session id
coral::ConnectionService::Guid coral::ConnectionService::ConnectionHandle::connectionId() const {
  return m_info->m_connectionId;
}

/// returns the service name associated to the connection
const std::string& coral::ConnectionService::ConnectionHandle::serviceName() const {
  return m_info->m_serviceName;
}

/// returns the technology name
const std::string& coral::ConnectionService::ConnectionHandle::technologyName() const {
  return m_info->m_technologyName;
}

/// returns the number of sessions
size_t coral::ConnectionService::ConnectionHandle::numberOfSessions() const {
  size_t nsess = 0;
  if(m_connection){
    nsess = dynamic_cast<coral::IDevConnection*>(m_connection.get())->size();
  }
  return nsess;
}

/// returns the server version
std::string coral::ConnectionService::ConnectionHandle::serverVersion() const{
  std::string sVer("");
  if( m_connection ) sVer = m_connection->serverVersion();
  return sVer;
}

/// returns the server version
coral::ITypeConverter& coral::ConnectionService::ConnectionHandle::typeConverter() const {
  if(!m_connection) throw coral::Exception("Session pointer is null","ConnectionService::ConnectionHandle::typeConverter","ConnectionService");
  return m_connection->typeConverter();
}


/// sets the start of idle time
void coral::ConnectionService::ConnectionHandle::setIdle(){
  m_info->m_startIdle = seal::Time::current();
  m_info->m_idle = true;
}

/// retrieves the idle flag
bool coral::ConnectionService::ConnectionHandle::isIdle() const {
  return m_info->m_idle;
}

/// returns true if timeout is reached
bool coral::ConnectionService::ConnectionHandle::isExpired() const {
  return (seal::Time::current()-m_info->m_startIdle).seconds() >= m_info->m_configuration.connectionTimeOut();
}

/// returns the specific time-out of the connection
int coral::ConnectionService::ConnectionHandle::specificTimeOut(){
  return m_info->m_configuration.connectionTimeOut();
}

