#include "ConnectionPool.h"
#include "ConnectionPoolTimer.h"
#include "ReplicaCatalogue.h"
#include "ConnectionServiceConfiguration.h"
#include "SealKernel/Context.h"
#include "SealKernel/MessageStream.h"
#include "SealBase/Time.h"
#include "RelationalAccess/ConnectionServiceException.h"
#include "RelationalAccess/SessionException.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/RelationalServiceException.h"
#include <vector>

/// constructor
coral::ConnectionService::ConnectionPool::ConnectionPool( const ConnectionServiceConfiguration& configuration ):
  m_idleConnections(),
  m_activeConnections(),
  m_connectionServiceConfiguration( configuration ),
  m_replicaCatalogue( new ReplicaCatalogue( configuration ) ),
  m_timerThread(0),
  m_finalized(false),
  m_mutexForPool(),
  m_mutexForThread(),
  m_poolStarted(false){
}

/// destructor
coral::ConnectionService::ConnectionPool::~ConnectionPool()
{
  delete m_replicaCatalogue;
  {
    boost::mutex::scoped_lock lock(m_mutexForThread);
    m_finalized = true;
  }
  if(m_timerThread)
  {
    m_timerThread->join();
    delete m_timerThread;
    m_timerThread = 0;
  }
}

/// returns the handle to a new constructed session, registered in the in-use session pool
coral::ConnectionService::SessionHandle
coral::ConnectionService::ConnectionPool::getSessionFromNewConnection(const ConnectionParams& connectionParameter,
                                                                      AccessMode accessMode){
  
  seal::IHandle<coral::IRelationalService> relationalService =
    m_connectionServiceConfiguration.relationalService();
  if(! relationalService ) throw ConnectionServiceException("None IRelationService instance is available",
                                                            "ConnectionPool::newConnection");

  if( m_connectionServiceConfiguration.isMonitoringEnabled() ) m_connectionServiceConfiguration.monitoringService();
  
  SessionHandle sessionHandle;

  seal::MessageStream log( m_connectionServiceConfiguration.serviceContext(),
                           m_connectionServiceConfiguration.serviceName(),
                           seal::Msg::Verbose );

  log << seal::Msg::Debug << "Trying to get a new "<<
    ((accessMode==coral::ReadOnly)? std::string("read-only") : std::string("update") )
      <<" session on \"" << connectionParameter.connectionString() <<"\""<< seal::flush;

  try {
    coral::IRelationalDomain& domain = relationalService->domainForConnection( connectionParameter.connectionString() );
    coral::IConnection* connection = domain.newConnection( connectionParameter.serviceName() ) ;
    if( connection ){

      ConnectionHandle connH( connection,
                              connectionParameter,
                              m_connectionServiceConfiguration );
      if(connH.open()){
        SessionHandle sessH( connH );
        sessH.open( connectionParameter.schemaName(),
                    connectionParameter.userName(),
                    connectionParameter.password(),
                    accessMode );
        sessionHandle = sessH;
      } else {
        m_replicaCatalogue->setConnectionExcluded( connectionParameter.technologyName(),
                                                   connectionParameter.serviceName() );
      }
    } else {
      log << seal::Msg::Debug << "Could not get a connection to service \"" << connectionParameter.serviceName() <<
        "\" from domain. "<< seal::flush;
      throw coral::ConnectionServiceException("Could not get a connection from domain.",
                                              "ConnectionPool::getSessionFromNewConnection");
    }
  } catch ( const coral::NonExistingDomainException& exc) {
    /// not clear what to do in this case? continue?
    log << seal::Msg::Warning << "Connection to \"" << connectionParameter.serviceName() <<
      "\" cannot be established: "<< exc.what()<< seal::flush;
  } catch (const coral::AuthenticationFailureException& exc){
    log << seal::Msg::Warning << "Authentication failure while attempting to start session for user=\""<<connectionParameter.userName()<<"\" on service \"" << connectionParameter.serviceName() << "\": "<< exc.what() << seal::flush;
  } catch (const coral::NoAuthenticationServiceException& exc){
    log << seal::Msg::Warning << "Authentication service missing - Required to start a session on connection \"" << connectionParameter.connectionString() << "\": "<< exc.what() << seal::flush;
  } catch (coral::MaximumNumberOfSessionsException& exc){
    log << seal::Msg::Warning << "Could not get a session on new connection to service \"" << connectionParameter.serviceName() <<"\": "<< exc.what() << seal::flush;
  } catch (const coral::ServerException& exc){
    log << seal::Msg::Warning << "Failure while attempting to connect to \"" << connectionParameter.connectionString() << "\": "<< exc.what() << seal::flush;
  }
  if( sessionHandle.isOpen() ){
    boost::mutex::scoped_lock lock(m_mutexForPool);
    m_activeConnections.add( sessionHandle.connection() );
    log << seal::Msg::Debug << " A New connection to service \""<<sessionHandle.connection().serviceName()<<"\" has been added in the active list. Id="<<sessionHandle.connection().connectionId()<< seal::flush;
  }  
  return sessionHandle; 
}



/// returns the handle to a re-used session from the pool, applying the subscribed policy.
coral::ConnectionService::SessionHandle
coral::ConnectionService::ConnectionPool::getSessionFromConnectionPool( const ConnectionParams& connectionParameter,
                                                                        AccessMode accessMode ){
  SessionHandle sessionHandle;
  
  seal::MessageStream log( m_connectionServiceConfiguration.serviceContext(),
                           m_connectionServiceConfiguration.serviceName(),
                           seal::Msg::Verbose);
  // if idle pool is not empty, first look up on it...
  if(m_idleConnections.size()>0){

    boost::mutex::scoped_lock lock(m_mutexForPool);
    ConnectionMap::Iterator iConn =
      m_idleConnections.lookup( connectionParameter.technologyName(),
                                connectionParameter.serviceName() );

    if( iConn!=m_idleConnections.end() ){
      try {
        if(iConn->isValid()){
          SessionHandle sessH(*iConn);
          sessH.open( connectionParameter.schemaName(),
                      connectionParameter.userName(),
                      connectionParameter.password(),
                      accessMode );
          if( sessH.isOpen() ){
            sessionHandle = sessH;
            m_idleConnections.remove( iConn );
            m_activeConnections.add( sessionHandle.connection() );
            log << seal::Msg::Debug << "The idle connection Id="<<sessionHandle.connection().connectionId()<<" to service \"" << sessionHandle.connection().serviceName() << "\" has been moved to the active list."<< seal::flush;
          }
        } else {
          log << seal::Msg::Debug << "Removing closed connection Id="<<iConn->connectionId()<<" on service=\""<<iConn->serviceName()<<"\" from the idle list."<< seal::flush;
          m_idleConnections.remove(iConn);
        }
      } catch (const coral::AuthenticationFailureException& exc){
        log << seal::Msg::Warning << "Authentication failure while attempting to start session for user=\""<<connectionParameter.userName()<<"\" on service \"" << connectionParameter.serviceName() << "\": "<< exc.what() << seal::flush;
      } catch (const coral::NoAuthenticationServiceException& exc){
        log << seal::Msg::Warning << "Authentication service missing - Required to start a session on connection \"" << connectionParameter.connectionString() << "\": "<< exc.what() << seal::flush;
      } catch (coral::MaximumNumberOfSessionsException& exc){
        log << seal::Msg::Warning << "Could not get a new session on connection to service \"" << iConn->serviceName() << "\" Id="<<iConn->connectionId()<<": "<< exc.what() << seal::flush;
      } catch (const coral::ServerException& exc){
        log << seal::Msg::Warning << "Failure while attempting to start a session on connection \"" << connectionParameter.connectionString() << "\": "<< exc.what() << seal::flush;
      }
    }
  }
  if( !sessionHandle && m_connectionServiceConfiguration.isConnectionSharingEnabled()){

    boost::mutex::scoped_lock lock(m_mutexForPool);
    ConnectionMap::Iterator iConn = m_activeConnections.lookup( connectionParameter.technologyName(),
                                                                connectionParameter.serviceName());
    if( iConn!=m_activeConnections.end() ){
      try {
        if(iConn->isValid()){

          SessionHandle sessH(*iConn);
          sessH.open( connectionParameter.schemaName(),
                      connectionParameter.userName(),
                      connectionParameter.password(),
                      accessMode );
          if( sessH.isOpen() ){
            sessionHandle = sessH;
            log << seal::Msg::Debug << "New session open on active connection Id="<<iConn->connectionId()<<" to service \"" << iConn->serviceName() <<"\""<< seal::flush;
          }
        } else {
          log << seal::Msg::Debug << "Removing closed connection Id="<<iConn->connectionId()<<" on service \""<<iConn->serviceName()<<"\" from the active list."<< seal::flush;
          m_activeConnections.remove(iConn);
        }
      } catch (const coral::AuthenticationFailureException& exc){
        log << seal::Msg::Warning << "Authentication failure while attempting to start session for user \""<<connectionParameter.userName()<<"\" on service \"" << connectionParameter.serviceName() << "\": "<< exc.what() << seal::flush;
      } catch (const coral::NoAuthenticationServiceException& exc){
        log << seal::Msg::Warning << "Authentication service missing - Required to start a session on connection \"" << connectionParameter.connectionString() << "\": "<< exc.what() << seal::flush;
      } catch (coral::MaximumNumberOfSessionsException& exc){
        log << seal::Msg::Warning << "Could not get a new session on connection to service \"" << iConn->serviceName() << "\" Id="<<iConn->connectionId()<<": "<< exc.what() << seal::flush;
      } catch (const coral::ServerException& exc){
        log << seal::Msg::Warning << "Failure while attempting to start a session on connection \"" << connectionParameter.connectionString() << "\": "<< exc.what() << seal::flush;
      }
    }
      
  }
  return sessionHandle;
}

/// combines the two above methods
coral::ConnectionService::SessionHandle
coral::ConnectionService::ConnectionPool::getValidSession( const std::string& connectionString,
                                                           const std::string& role,
                                                           AccessMode accessMode ){
  // initialize pool if required
  startPool();
  
  seal::MessageStream log( m_connectionServiceConfiguration.serviceContext(),
                           m_connectionServiceConfiguration.serviceName(),
                           seal::Msg::Verbose );

  SessionHandle sessionHandle;
  
  std::vector<ConnectionParams>* replicas = m_replicaCatalogue->getReplicas(connectionString, role, accessMode);
  
  unsigned int iReplica = 0;
  bool sessionOk = false;
  while( iReplica < replicas->size() && !sessionOk ){
    sessionHandle = getSessionFromConnectionPool( (*replicas)[iReplica], accessMode );
    sessionOk = sessionHandle.isOpen();
    if( !sessionOk ) {
      log << seal::Msg::Debug << "No valid connection for string \""<< connectionString << "\" found in the pool. Creating a new connection." << seal::flush;
      if(!m_replicaCatalogue->isConnectionExcluded((*replicas)[iReplica].technologyName(),(*replicas)[iReplica].serviceName())){
        sessionHandle = getSessionFromNewConnection( (*replicas)[iReplica], accessMode );
        sessionOk = sessionHandle.isOpen();
      }
    }
    iReplica++;
  }
  delete replicas;
  if(!sessionOk){
    throw coral::ConnectionNotAvailableException( connectionString,"ConnectionPool::getSessionFromNewConnection" );
  }
  
  return sessionHandle;
}

/// releases the use of the specified session
bool
coral::ConnectionService::ConnectionPool::releaseConnection( ConnectionHandle& connection ){
  if(!connection) return false;

  boost::mutex::scoped_lock lock(m_mutexForPool);
  bool connected = connection.isValid();
  unsigned nsess = connection.numberOfSessions();
  seal::MessageStream log( m_connectionServiceConfiguration.serviceContext(),
                           m_connectionServiceConfiguration.serviceName(),
                           seal::Msg::Verbose );
  log << seal::Msg::Debug << "Connection id="<<connection.connectionId()<<" to service \""
            <<connection.serviceName() << "\" has "<<nsess<<" session(s) opended."<< seal::flush; 
  if(!connected || nsess==0 ){
    m_activeConnections.remove( connection );
    log << seal::Msg::Debug << "The active connection id="<<connection.connectionId() <<" to service \""
            <<connection.serviceName() << "\" has been removed from the active list." << seal::flush;
    if(connected) {
      connection.setIdle();
      // if time out is set>0, move the idle connection to idle pool
      if(connection.specificTimeOut()>0){
        m_idleConnections.add( connection );
        log << seal::Msg::Debug << "The valid active connection id="<<connection.connectionId() <<" to service \""
            <<connection.serviceName() << "\" has been moved to the idle list." << seal::flush;
      }
    } else {
      log << seal::Msg::Debug << "The invalid connection id="<<connection.connectionId()<<" to service \""
          <<connection.serviceName() << "\" has been removed from pool." << seal::flush;
    }
    
  }
  return true;
}

/// releases the use of the specified session
bool
coral::ConnectionService::ConnectionPool::removeConnection( ConnectionHandle& connection ){
  if( !connection ) return false;
  boost::mutex::scoped_lock lock(m_mutexForPool);
  m_activeConnections.remove( connection );
  seal::MessageStream log( m_connectionServiceConfiguration.serviceContext(),
                           m_connectionServiceConfiguration.serviceName(),
                           seal::Msg::Verbose);
  log << seal::Msg::Debug << "The invalid connection id="<< connection.connectionId()<<" to service \""
      <<connection.serviceName() << "\" has been removed from pool." << seal::flush;
  return true;
}

/// close all the sessions of the pool
void
coral::ConnectionService::ConnectionPool::clearConnectionPool() {
  boost::mutex::scoped_lock lock(m_mutexForPool);
  m_idleConnections.clear();
  m_activeConnections.clear();
}

/// remove from the pool the session timed out.
void
coral::ConnectionService::ConnectionPool::cleanUpTimedOutConnections()
{
  boost::mutex::scoped_lock lockForDestruction(m_mutexForThread);
  boost::mutex::scoped_lock lockForCleanUp(m_mutexForPool);
  if( m_idleConnections.size()>0){
    std::vector< ConnectionMap::Iterator > timedOutConnections;
  
    for( ConnectionMap::Iterator iter = m_idleConnections.begin(); iter!= m_idleConnections.end(); ++iter ){
      if( iter->isExpired() ){
        timedOutConnections.push_back( iter );
      }
    }
  
    if(!m_finalized){
      seal::MessageStream log(   m_connectionServiceConfiguration.serviceContext(),
                                 m_connectionServiceConfiguration.serviceName(),
                                 seal::Msg::Verbose);
      log << seal::Msg::Debug << "Checking idle pool for timed-out connections - Total="<<m_idleConnections.size()<<", Expired="<<timedOutConnections.size()  << seal::flush;
    }
  
    for( std::vector< ConnectionMap::Iterator >::iterator vecIter = timedOutConnections.begin();
         vecIter!= timedOutConnections.end();
         vecIter++ ){
      if(!m_finalized){
        seal::MessageStream log(   m_connectionServiceConfiguration.serviceContext(),
                                   m_connectionServiceConfiguration.serviceName(),
                                   seal::Msg::Verbose);
        log << seal::Msg::Debug << "The timed-out idle connection id="<<(*vecIter)->connectionId()<<" to service \""<<(*vecIter)->serviceName() << "\" has been removed from pool." << seal::flush;
      }
      m_idleConnections.remove( *vecIter );
    }
  }
}

/// returns the number of idle session in the pool
size_t
coral::ConnectionService::ConnectionPool::numberOfIdleConnections(){
  boost::mutex::scoped_lock lock(m_mutexForPool);
  return m_idleConnections.size();
}


/// returns the number of active session in the pool
size_t
coral::ConnectionService::ConnectionPool::numberOfActiveConnections(){
  boost::mutex::scoped_lock lock(m_mutexForPool);
  return m_activeConnections.size();
}

/// returns true if the pool cleaner thread is requested to stop
bool coral::ConnectionService::ConnectionPool::isFinalized(){
  boost::mutex::scoped_lock lock(m_mutexForThread);
  return m_finalized;
}

void coral::ConnectionService::ConnectionPool::startPool(){
  boost::mutex::scoped_lock lock(m_mutexForPool);
  if(!m_poolStarted){
    m_poolStarted = true;
    if(m_connectionServiceConfiguration.connectionTimeOut()>0 &&
       m_connectionServiceConfiguration.isPoolAutomaticCleanUpEnabled()){
      // instantiate cleaner timer
      ConnectionPoolTimer timer(*this,m_connectionServiceConfiguration.poolCleanUpPeriod());
      m_timerThread = new boost::thread(timer);
    }
  }
}


