#include "ConnectionService.h"
#include "ConnectionServiceConfiguration.h"
#include "ConnectionPool.h"
#include "SessionProxy.h"
#include "WebCacheControl.h"
#include "SealKernel/Context.h"
#include "SealKernel/PropertyManager.h"
#include "CoralBase/Exception.h"

class CallbackForParameter
{
  public:
    explicit CallbackForParameter( coral::ConnectionService::ConnectionServiceConfiguration& configuration )
      : m_configuration( configuration )
    {
    }

    void operator() ( const seal::PropertyBase& p )
    {
      int period = boost::any_cast< int >( p.get() );
      m_configuration.poolCleanUpPeriod() = period;
    }

  private:
    coral::ConnectionService::ConnectionServiceConfiguration& m_configuration;
};

DEFINE_SEAL_COMPONENT ( coral::ConnectionService::ConnectionService,"CORAL/Services/ConnectionService")
  
coral::ConnectionService::ConnectionService::ConnectionService( seal::Context* ctx )
  :seal::Service( ctx, classContextKey() ),
   m_connectionPool( 0 ),
   m_configuration( 0 ),
   m_webCacheControl( 0 ),
   m_sessions(),
   m_mutexLock(){
  m_configuration = new ConnectionServiceConfiguration( ctx, seal::Service::name() );
  propertyManager()->declareProperty( "PoolCleanUpPeriod",
                                      m_configuration->poolCleanUpPeriod(),
                                      10,
                                      "The period of the connection pool clean up thread",
                                      CallbackForParameter( *m_configuration ) );
  m_connectionPool = new ConnectionPool( *m_configuration );
  m_webCacheControl =  new WebCacheControl( ctx, seal::Service::name() );
}

coral::ConnectionService::ConnectionService::ConnectionService( seal::Context* ctx, const std::string& key )
  : seal::Service( ctx, key ),
    m_connectionPool( 0 ),
    m_configuration( 0 ),
    m_webCacheControl( 0 ),
    m_sessions(){
  m_configuration = new ConnectionServiceConfiguration( ctx, seal::Service::name() );
  propertyManager()->declareProperty( "PoolCleanUpPeriod",
                                      m_configuration->poolCleanUpPeriod(),
                                      10,
                                      "The period of the connection pool clean up thread",
                                      CallbackForParameter( *m_configuration ) );
  m_connectionPool = new ConnectionPool( *m_configuration );
  m_webCacheControl = new WebCacheControl ( ctx, seal::Service::name() );
}

coral::ConnectionService::ConnectionService::~ConnectionService()
{
  for( std::set<SessionProxy*>::const_iterator iSess = m_sessions.begin(); iSess!=m_sessions.end();iSess++)
  {
    (*iSess)->invalidate();
  }
  delete m_connectionPool;
  delete m_configuration;
  delete m_webCacheControl;
}

coral::ISessionProxy* coral::ConnectionService::ConnectionService::connect( const std::string& connectionString,
                                                                            AccessMode accessMode  ) 
{
  SessionProxy* theSession = 0;
  try {
    boost::mutex::scoped_lock lock(m_mutexLock);
    theSession = new coral::ConnectionService::SessionProxy( connectionString, accessMode, this );
    theSession->open();
    m_sessions.insert(theSession);
    m_connectionPool->cleanUpTimedOutConnections();
  } catch (const coral::Exception& exc){
    delete theSession;
    theSession = 0;
    throw;
  }
  return static_cast<ISessionProxy*>(theSession);
}

coral::ISessionProxy* coral::ConnectionService::ConnectionService::connect( const std::string& connectionString,
                                                                            const std::string& asRole,
                                                                            AccessMode accessMode  ) 
{
  SessionProxy* theSession = 0;
  try {
    boost::mutex::scoped_lock lock(m_mutexLock);
    theSession = new coral::ConnectionService::SessionProxy( connectionString, asRole, accessMode, this );
    theSession->open();
    m_sessions.insert(theSession);
    m_connectionPool->cleanUpTimedOutConnections();
  } catch (const coral::Exception& exc){
    delete theSession;
    theSession = 0;
    throw;
  }
  return static_cast<ISessionProxy*>(theSession);
}

void
coral::ConnectionService::ConnectionService::purgeConnectionPool(){
  boost::mutex::scoped_lock lock(m_mutexLock);
  m_connectionPool->cleanUpTimedOutConnections();  
}


coral::IConnectionServiceConfiguration& coral::ConnectionService::ConnectionService::configuration() 
{
  return *m_configuration;
}

/**
 * Returns the monitoring reporter
 */
const coral::IMonitoringReporter& coral::ConnectionService::ConnectionService::monitoringReporter() const
{
  return m_configuration->monitoringService()->reporter();
}

coral::IWebCacheControl& coral::ConnectionService::ConnectionService::webCacheControl()
{
  return *m_webCacheControl;
}

/// returns the reference to the owned Connection Pool
coral::ConnectionService::ConnectionPool& coral::ConnectionService::ConnectionService::connectionPool(){
  return *m_connectionPool;
}

/// removes a session from the registry
void
coral::ConnectionService::ConnectionService::unRegisterSession( SessionProxy* sessionProxy ){
  boost::mutex::scoped_lock lock(m_mutexLock);
  m_sessions.erase(sessionProxy);
}


/// returns the number of idle session in the pool
size_t coral::ConnectionService::ConnectionService::numberOfIdleConnectionsInPool(){
  return m_connectionPool->numberOfIdleConnections();
}


/// returns the number of active session in the pool
size_t coral::ConnectionService::ConnectionService::numberOfActiveConnectionsInPool(){
  return m_connectionPool->numberOfActiveConnections();
}


