#include "ConnectionServiceConfiguration.h"
#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/IReplicaSortingAlgorithm.h"
#include "RelationalAccess/ConnectionServiceException.h"
#include "SealKernel/PropertyManager.h"
#include <stdexcept>
#include "PluginManager/Module.h"
#include "SealKernel/ComponentFactory.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/MessageStream.h"
#include <sstream>

/// destructor
coral::ConnectionService::ConnectionServiceConfiguration::~ConnectionServiceConfiguration(){
  if(m_sortingAlgorithm) delete m_sortingAlgorithm;
}

/// constructor
coral::ConnectionService::ConnectionServiceConfiguration::ConnectionServiceConfiguration( seal::Context* context,
                                                                                          const std::string& serviceName ):
  m_serviceContext( context ),
  m_serviceName( serviceName ),
  m_replicaFailover( true ),
  m_connectionSharing( true ),
  m_readOnlyOnUpdate( true ),
  m_automaticCleanUp( true ),
  m_periodOfConnectionRetrials( 10 ),
  m_connectionRetrialTimeOut( 60 ),
  m_connectionTimeOut( 300 ),
  m_exclusionTime( 3600 ),
  m_monitoringLevel( coral::monitor::Off ),
  m_defaultAuthenticationService(""), 
  m_defaultLookupService( "CORAL/Services/XMLLookupService" ),
  m_defaultMonitoringService( "CORAL/Services/MonitoringService" ),
  m_defaultRelationalService( "CORAL/Services/RelationalService" ),
  m_sortingAlgorithm(0),
  m_poolCleanUpPeriod( 10 ){
  
  char* authPath = ::getenv ( "CORAL_AUTH_PATH" );
  if ( authPath ) {
    m_defaultAuthenticationService = "CORAL/Services/XMLAuthenticationService";
  }
  else {
    if ( ::getenv ( "CORAL_AUTH_USER" ) && ::getenv ( "CORAL_AUTH_PASSWORD" ) )
      m_defaultAuthenticationService = "CORAL/Services/EnvironmentAuthenticationService";
    else
      m_defaultAuthenticationService = "CORAL/Services/XMLAuthenticationService";
  }
  
}

/// loads the required plugins according to defaults
seal::IHandle<coral::IRelationalService>
coral::ConnectionService::ConnectionServiceConfiguration::relationalService() const {
  std::vector< seal::IHandle< coral::IRelationalService > > loadedServices;
  m_serviceContext->query( loadedServices );
  if ( loadedServices.empty() ) {
    seal::MessageStream log( m_serviceContext, m_serviceName );
    log << seal::Msg::Info << "Loading default plugin for coral::IRelationalService: "
        <<m_defaultRelationalService << seal::flush;
    loadService( m_defaultRelationalService );
    m_serviceContext->query( loadedServices );
    if(loadedServices.empty() ) {
      std::ostringstream mess;
      mess << "Default plugin "<< m_defaultRelationalService <<" has been not found.";
      log << seal::Msg::Error << mess.str() << seal::flush;
      throw ConnectionServiceException(mess.str(),
                                       "ConnectionService::ConnectionServiceConfiguration::relationalService");
    }                                                            
  }
  return loadedServices.front();
}

/// loads the required plugins according to defaults
seal::IHandle<coral::IAuthenticationService>
coral::ConnectionService::ConnectionServiceConfiguration::authenticationService() const {
  std::vector< seal::IHandle< coral::IAuthenticationService > > loadedServices;
  m_serviceContext->query( loadedServices );
  if ( loadedServices.empty() ) {
    seal::MessageStream log( m_serviceContext, m_serviceName );
    log << seal::Msg::Info << "Loading default plugin for coral::IAuthenticationService: "
        <<m_defaultAuthenticationService << seal::flush;    
    loadService( m_defaultAuthenticationService );
    m_serviceContext->query( loadedServices );
    if(loadedServices.empty() ) {
      std::ostringstream mess;
      mess << "Default plugin "<< m_defaultAuthenticationService <<" has been not found.";
      log << seal::Msg::Error << mess.str() << seal::flush;
      throw ConnectionServiceException(mess.str(),
                                       "ConnectionService::ConnectionServiceConfiguration::authenticationService");
    }                                                            
  }
  return loadedServices.front();
}

/// loads the required plugins according to defaults
seal::IHandle<coral::ILookupService>
coral::ConnectionService::ConnectionServiceConfiguration::lookupService() const {
  std::vector< seal::IHandle< coral::ILookupService > > loadedServices;
  m_serviceContext->query( loadedServices );
  if ( loadedServices.empty() ) {
    seal::MessageStream log( m_serviceContext, m_serviceName );
    log << seal::Msg::Info << "Loading default plugin for coral::ILookupService: "
        <<m_defaultLookupService << seal::flush;    
    loadService( m_defaultLookupService );
    m_serviceContext->query( loadedServices );
    if(loadedServices.empty() ) {
      std::ostringstream mess;
      mess << "Default plugin "<< m_defaultLookupService <<" has been not found.";
      log << seal::Msg::Error << mess.str() << seal::flush;
      throw ConnectionServiceException(mess.str(),
                                       "ConnectionService::ConnectionServiceConfiguration::lookupService");
    }                                                            
  }
  return loadedServices.front();
}

/// loads the required plugin according to defaults
seal::IHandle<coral::monitor::IMonitoringService>
coral::ConnectionService::ConnectionServiceConfiguration::monitoringService() const {
  std::vector< seal::IHandle< coral::monitor::IMonitoringService > > loadedServices;
  m_serviceContext->query( loadedServices );
  if ( loadedServices.empty() ) {
    seal::MessageStream log( m_serviceContext, m_serviceName );
    log << seal::Msg::Info << "Loading default plugin for coral::IMonitoringService: "
        <<m_defaultMonitoringService << seal::flush;
    loadService( m_defaultMonitoringService );
    m_serviceContext->query( loadedServices );
    if(loadedServices.empty() ) {
      std::ostringstream mess;
      mess << "Default plugin "<< m_defaultMonitoringService <<" has been not found.";
      log << seal::Msg::Error << mess.str() << seal::flush;
      throw ConnectionServiceException(mess.str(),
                                       "ConnectionService::ConnectionServiceConfiguration::monitoringService");
    }                                                            
  }
  return loadedServices.front();
}

void
coral::ConnectionService::ConnectionServiceConfiguration::setReplicaSortingAlgorithm(coral::IReplicaSortingAlgorithm* algorithm){
  if(m_sortingAlgorithm && m_sortingAlgorithm!=algorithm) {
    delete  m_sortingAlgorithm;
    m_sortingAlgorithm = 0;
  }
  m_sortingAlgorithm = algorithm;
}


/// access to the service Context
seal::Context*
coral::ConnectionService::ConnectionServiceConfiguration::serviceContext() const {
  return m_serviceContext.get();
}

/// access to the upper levevl service name
const std::string&
coral::ConnectionService::ConnectionServiceConfiguration::serviceName() const {
  return m_serviceName;
}

/// loads the the specified Service
void
coral::ConnectionService::ConnectionServiceConfiguration::loadService( const std::string& serviceName ) const{
  // Load the default RelationalService into the top-level context
  seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( m_serviceContext.get() );
  loader->load( serviceName);
}

