#ifndef CONNECTIONSERVICE_CONNECTIONSERVICECONFIGURATION_H
#define CONNECTIONSERVICE_CONNECTIONSERVICECONFIGURATION_H

#include "RelationalAccess/IAuthenticationService.h"
#include "RelationalAccess/IConnectionServiceConfiguration.h"
#include "RelationalAccess/ILookupService.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/IRelationalService.h"

#include "SealKernel/Context.h"

namespace coral {

  namespace ConnectionService{

    /// implementation of the interface IConnectionServiceConfiguration
    class ConnectionServiceConfiguration : virtual public IConnectionServiceConfiguration {
      public:  

      /// constructor
      ConnectionServiceConfiguration( seal::Context* context, const  std::string& serviceName );

      /// destructor
      virtual ~ConnectionServiceConfiguration();

      /**
       * Enable the failing over to the next available
       * replica in case the current one is not available.
       * Otherwise the ConnectionService gives up.
       */
      void enableReplicaFailOver();
   
      /**
       * Disable the failing over to the next available
       * replica in case the current one is not available.
       */
      void disableReplicaFailOver();
    
      /**
       * Returns the failover mode
       */
      bool isReplicaFailoverEnabled() const;
      
      /**
       * Enables the sharing of the same physical connection
       * among more clients.
       */
      void enableConnectionSharing();

      /**
       * Disables the sharing of the same physical connection
       * among more clients.
       */
      void disableConnectionSharing();

      /**
       * Returns true if the connction sharing is enabled 
       */
      bool isConnectionSharingEnabled() const;
      
      /**
       * Enables the re-use of Update connections for Read-Only sessions
       */
      void enableReadOnlySessionOnUpdateConnections();

      /**
       * Disables the re-use of Update connections for Read-Only sessions
       */
      void disableReadOnlySessionOnUpdateConnections();

      /**
       * Returns true if the  re-use of Update connections for Read-Only sessions is enabled 
       */
      bool isReadOnlySessionOnUpdateConnectionsEnabled() const;

      /**
       * Sets the period of connection retrials (time interval between two retrials).
       */
      void setConnectionRetrialPeriod( int timeInSeconds );

      /**
       * Returns the rate of connection retrials (time interval between two retrials).
       */
      int connectionRetrialPeriod() const;

      /**
       * Sets the time out for the connection retrials before the connection
       * service fails over to the next available replica or quits.
       */
      void setConnectionRetrialTimeOut( int timeOutInSeconds );

      /**
       * Returns the time out for the connection retrials before the connection
       * service fails over to the next available replica or quits.
       */
      int connectionRetrialTimeOut() const;
      
      /**
       * Sets the connection time out in seconds.
       */
      void setConnectionTimeOut( int timeOutInSeconds );    

      /**
       * Retrieves the connection time out in seconds.
       */
      int connectionTimeOut() const;

     /**
      * Activate the parallel thread for idle pool cleaning up
      */
     void enablePoolAutomaticCleanUp();
    
     /**
      * Disable the parallel thread for idle pool cleaning up
      */
     void disablePoolAutomaticCleanUp();

     /**
      * Returns true if the parallel thread for idle pool cleaning up is enabled
      */
     bool isPoolAutomaticCleanUpEnabled() const;

      /**
       * Sets the time duration of exclusion from failover list for a
       * connection not available.
       */
      void setMissingConnectionExclusionTime( int timeInSeconds );    

      /**
       * Retrieves the time duration of exclusion from failover list for a
       * connection not available.
       */
      int missingConnectionExclusionTime() const;

    /**
       * Sets the monitoring level for the new sessions.
       */
      void setMonitoringLevel( coral::monitor::Level level );

      /**
       * Retrieves the current monitoring level.
       */
      coral::monitor::Level monitoringLevel() const ;

      /**
       * Loads and sets the default authentication service to be used for the new sessions.
       */
      void setDefaultAuthenticationService( const std::string serviceName);

      /**
       * Loads and sets the default lookup service to be used for the new sessions.
       */
      void setDefaultLookupService( const std::string& serviceName );
  
      /**
       * Loads and sets the default relational service to be used for the new sessions.
       */
      void setDefaultRelationalService( const std::string& serviceName );
      
      /**
       * Loads and sets the default monitoring service to be used for the new sessions.
       */
      void setDefaultMonitoringService( const std::string& serviceName );

      /**
       * Sets the algorithm to be used for the database replica ordering.
       */
      void setReplicaSortingAlgorithm(coral::IReplicaSortingAlgorithm* algorithm);

      /// returns true if the monitoring level has been set
      bool isMonitoringEnabled() const;
      
      /// loads the required plugin according to defaults
      seal::IHandle<coral::IRelationalService> relationalService() const;
      
      /// loads the required plugin according to defaults
      seal::IHandle<coral::IAuthenticationService> authenticationService() const;

      /// loads the required plugin according to defaults
      seal::IHandle<coral::ILookupService> lookupService() const;

      /// loads the required plugin according to defaults
      seal::IHandle<coral::monitor::IMonitoringService> monitoringService() const;

      coral::IReplicaSortingAlgorithm* sortingAlgorithm() const;
      
      /// access to the service Context
      seal::Context* serviceContext() const;

      /// upper level service name
      const std::string& serviceName() const;

      /// access to the pool clean up period
      const int& poolCleanUpPeriod() const;
      int& poolCleanUpPeriod();
 
      private:

      /// loads the the specified Service 
      void loadService( const std::string& serviceName ) const;
      
      private:

      /// the service context
      seal::Handle<seal::Context> m_serviceContext;

      /// the upper level service name
      std::string m_serviceName;
      
      /// the flag enabling replica failover
      bool m_replicaFailover;

      /// the flag enabling connection sharing
      bool m_connectionSharing;

      /// the flag enabling Read-Only re-use of Update connection
      bool m_readOnlyOnUpdate;

      /// the flag enabling the automatic clean up
      bool m_automaticCleanUp;
      
      /// the number of connection retrials per minute
      int m_periodOfConnectionRetrials;

      /// the time out for the connection retrial
      int m_connectionRetrialTimeOut;

      /// the life time of the unused connections
      int m_connectionTimeOut;

      /// the time of the exclusion of unavailable connection
      int m_exclusionTime;

      /// the stored monitoring level
      coral::monitor::Level m_monitoringLevel;

      /// the name of the default authentication service
      std::string m_defaultAuthenticationService;

      /// the name of the default lookup service
      std::string m_defaultLookupService;

      /// the name of the default monitoring service
      std::string m_defaultMonitoringService;

      /// the name of the default relational service
      std::string m_defaultRelationalService;

      coral::IReplicaSortingAlgorithm* m_sortingAlgorithm;

      /// the period of the pool clean up
      int m_poolCleanUpPeriod;

    };
  }
  
}   				

/**
 * Enable the failing over to the next available
 * replica in case the current one is not available.
 * Otherwise the ConnectionService gives up.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::enableReplicaFailOver(){
  m_replicaFailover = true;
}

/**
 * Disable the failing over to the next available
 * replica in case the current one is not available.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::disableReplicaFailOver(){
  m_replicaFailover = false;
}

/**
 * Returns the failover mode
 */
inline
bool
coral::ConnectionService::ConnectionServiceConfiguration::isReplicaFailoverEnabled() const {
  return m_replicaFailover;
}

/**
 * Enables the sharing of the same physical connection
 * among more clients.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::enableConnectionSharing(){
  m_connectionSharing = true;
}

/**
 * Disables the sharing of the same physical connection
 * among more clients.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::disableConnectionSharing(){
  m_connectionSharing = false;
}

/**
 * Returns true if the connction sharing is enabled 
 */
inline
bool
coral::ConnectionService::ConnectionServiceConfiguration::isConnectionSharingEnabled() const {
  return m_connectionSharing;
}

/**
 * Enables the re-use of Update connections for Read-Only sessions
 */
inline
void 
coral::ConnectionService::ConnectionServiceConfiguration::enableReadOnlySessionOnUpdateConnections(){
  m_readOnlyOnUpdate = true;
}
      
/**
 * Disables the re-use of Update connections for Read-Only sessions
 */
inline
void 
coral::ConnectionService::ConnectionServiceConfiguration::disableReadOnlySessionOnUpdateConnections(){
  m_readOnlyOnUpdate = false;
}

/**
 * Returns true if the  re-use of Update connections for Read-Only sessions is enabled 
 */
inline
bool 
coral::ConnectionService::ConnectionServiceConfiguration::isReadOnlySessionOnUpdateConnectionsEnabled() const {
  return m_readOnlyOnUpdate;
}
    
/**
 * Sets the period of connection retrials (time interval between two retrials).
 */
inline
void 
coral::ConnectionService::ConnectionServiceConfiguration::setConnectionRetrialPeriod( int timeInSeconds ){
  m_periodOfConnectionRetrials = timeInSeconds;
}

/**
 * Returns the rate of connection retrials (time interval between two retrials).
 */
inline
int 
coral::ConnectionService::ConnectionServiceConfiguration::connectionRetrialPeriod() const {
  return m_periodOfConnectionRetrials;
}

/**
 * Sets the time out for the connection retrials before the connection
 * service fails over to the next available replica or quits.
 */
inline 
void 
coral::ConnectionService::ConnectionServiceConfiguration::setConnectionRetrialTimeOut( int timeOutInSeconds ){
  m_connectionRetrialTimeOut = timeOutInSeconds; 
}

/**
 * Returns the time out for the connection retrials before the connection
 * service fails over to the next available replica or quits.
 */
inline 
int 
coral::ConnectionService::ConnectionServiceConfiguration::connectionRetrialTimeOut() const {
  return m_connectionRetrialTimeOut;
}
      
/**
 * Sets the connection time out in seconds.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::setConnectionTimeOut( int timeOutInSeconds ){
  m_connectionTimeOut = timeOutInSeconds;
}

/**
 * Retrieves the connection time out in seconds.
 */
inline
int
coral::ConnectionService::ConnectionServiceConfiguration::connectionTimeOut() const {
  return m_connectionTimeOut;
}

/**
 * Activate the parallel thread for idle pool cleaning up
 */
inline 
void
coral::ConnectionService::ConnectionServiceConfiguration::enablePoolAutomaticCleanUp(){
  m_automaticCleanUp = true;
}

/**
 * Disable the parallel thread for idle pool cleaning up
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::disablePoolAutomaticCleanUp(){
  m_automaticCleanUp = false;  
}

/**
 * Returns true if the parallel thread for idle pool cleaning up is enabled
 */
inline
bool
coral::ConnectionService::ConnectionServiceConfiguration::isPoolAutomaticCleanUpEnabled() const {
  return m_automaticCleanUp;
}

/**
 * Sets the time duration of exclusion from failover list for a
 * connection not available.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::setMissingConnectionExclusionTime( int timeInSeconds )
{
  m_exclusionTime = timeInSeconds;
}

/**
 * Retrieves the time duration of exclusion from failover list for a
 * connection not available.
 */
inline
int
coral::ConnectionService::ConnectionServiceConfiguration::missingConnectionExclusionTime() const {
  return m_exclusionTime;
}

/**
 * Sets the monitoring level for the new sessions.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::setMonitoringLevel( coral::monitor::Level level ){
  m_monitoringLevel = level;
}

/**
 * Retrieves the current monitoring level.
 */
inline
coral::monitor::Level
coral::ConnectionService::ConnectionServiceConfiguration::monitoringLevel() const {
  return m_monitoringLevel;
}

/**
 * Loads and sets the default authentication service to be used for the new sessions.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::setDefaultAuthenticationService( const std::string serviceName) {
  m_defaultAuthenticationService = serviceName;
}

/**
 * Loads and sets the default lookup service to be used for the new sessions.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::setDefaultLookupService( const std::string& serviceName ){
  m_defaultLookupService = serviceName;
}

/**
 * Loads and sets the default relational service to be used for the new sessions.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::setDefaultRelationalService( const std::string& serviceName ){
  m_defaultRelationalService = serviceName;
}     

/**
 * Loads and sets the default monitoring service to be used for the new sessions.
 */
inline
void
coral::ConnectionService::ConnectionServiceConfiguration::setDefaultMonitoringService( const std::string& serviceName ){
  m_defaultMonitoringService = serviceName;
}

inline
coral::IReplicaSortingAlgorithm*
coral::ConnectionService::ConnectionServiceConfiguration::sortingAlgorithm() const {
  return m_sortingAlgorithm;
}
      
/// returns true if the monitoring level has been set
inline 
bool 
coral::ConnectionService::ConnectionServiceConfiguration::isMonitoringEnabled() const {
  return m_monitoringLevel != coral::monitor::Off;
}

/// access to the pool clean up period
inline 
const int& 
coral::ConnectionService::ConnectionServiceConfiguration::poolCleanUpPeriod() const {
  return m_poolCleanUpPeriod;
}
inline 
int& 
coral::ConnectionService::ConnectionServiceConfiguration::poolCleanUpPeriod() {
  return m_poolCleanUpPeriod;
}

#endif  
