#include "ReplicaCatalogue.h"
#include "ConnectionString.h"
#include "RelationalAccess/ConnectionServiceException.h"
#include "ConnectionServiceConfiguration.h"
#include "SealKernel/Context.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/MessageStream.h"
#include "SealBase/Time.h"
#include "RelationalAccess/AuthenticationServiceException.h"
#include "RelationalAccess/ILookupService.h"
#include "RelationalAccess/IDatabaseServiceSet.h"
#include "RelationalAccess/IDatabaseServiceDescription.h"
#include "RelationalAccess/IAuthenticationService.h"
#include "RelationalAccess/IAuthenticationCredentials.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/IReplicaSortingAlgorithm.h"

/// constructor
coral::ConnectionService::ReplicaCatalogue::ReplicaCatalogue( const ConnectionServiceConfiguration& serviceConfiguration ):
  m_serviceConfiguration( serviceConfiguration ),
  m_excludedConnections(),
  m_mutexForExclusionList(){
}

/// destructor
coral::ConnectionService::ReplicaCatalogue::~ReplicaCatalogue(){}

/// sets the specified connection as unavailable
void
coral::ConnectionService::ReplicaCatalogue::setConnectionExcluded( const std::string& technologyName,
                                                                   const std::string& serviceName ){
  seal::MessageStream log( m_serviceConfiguration.serviceContext(),
                           m_serviceConfiguration.serviceName(),
                           seal::Msg::Verbose );
  {
    boost::mutex::scoped_lock lock(m_mutexForExclusionList);
    std::pair<std::map<std::string,seal::Time>::iterator,bool> ret =
      m_excludedConnections.insert(
        std::make_pair(ConnectionString::serviceKey(technologyName,serviceName), seal::Time::current()) );
    if(!ret.second) ret.first->second = seal::Time::current();
  }
  log << seal::Msg::Warning << "Service " << serviceName << " (technology=\""<<technologyName<<"\") is currently unavailable, and will be temporarely excluded from replicas." << seal::flush;
}

/// resolves the technology specified in the connection string, checking if it is supported by an
/// existing plug-in...
bool
coral::ConnectionService::ReplicaCatalogue::isConnectionPhysical( const std::string& connectionString ) const {
  bool physical = false;
  seal::IHandle<coral::IRelationalService> relationalService = m_serviceConfiguration.relationalService();
  if(! relationalService ) throw ConnectionServiceException("None IRelationService instance is available",
                                                            "ReplicaCatalogue::isConnectionPhysical"  );
  std::vector<std::string> technologies = relationalService->availableTechnologies();
  std::vector<std::string>::iterator iter=technologies.begin();
  while( (iter != technologies.end()) && (physical == false)) {
    std::string::size_type pos = connectionString.find(*iter);
    if (pos != std::string::npos){
      physical = true;
    }
    iter++;
  }
  return physical;
}
      
/// resolves the technology specified in the connection string, checking if it is supported by an
/// existing plug-in...
bool
coral::ConnectionService::ReplicaCatalogue::isConnectionExcluded( const std::string& technologyName,
                                                                  const std::string& serviceName ) const {
  boost::mutex::scoped_lock lock(m_mutexForExclusionList);
  bool excluded = false;
  std::map<std::string,seal::Time>::iterator iConn =
    m_excludedConnections.find( ConnectionString::serviceKey(technologyName,serviceName ));
  if(iConn!=m_excludedConnections.end()) {
    seal::MessageStream log( m_serviceConfiguration.serviceContext(),
                             m_serviceConfiguration.serviceName(),
                             seal::Msg::Verbose );
    excluded = (( seal::Time::current().second(true)-iConn->second.second(true)) < m_serviceConfiguration.missingConnectionExclusionTime() );
    if(!excluded) {
      m_excludedConnections.erase(iConn);
      log << seal::Msg::Debug << "Service " << serviceName << " (technology=\""<<technologyName<<"\") has been re-admitted in the list of replicas." << seal::flush;      
    } else {
      log << seal::Msg::Info << "Service " << serviceName << " (technology=\""<<technologyName<<"\") is currently excluded from replicas." << seal::flush;  
    }
  }
  return excluded;
}

/// returns the credentials for the specified connection string
std::pair<std::string,std::string>*
coral::ConnectionService::ReplicaCatalogue::getCredentials( const std::string& connectionString,
                                                            const std::string& role ) const {
  std::pair<std::string, std::string>* credentials =  new std::pair<std::string,std::string>("","");
  seal::IHandle<coral::IAuthenticationService> authenticationService = m_serviceConfiguration.authenticationService();
  seal::MessageStream log( m_serviceConfiguration.serviceContext(),
                           m_serviceConfiguration.serviceName(),
                           seal::Msg::Verbose );
  if( authenticationService )
  {
    try {
      if(role.empty()){
        credentials->first = authenticationService->credentials( connectionString ).valueForItem( IAuthenticationCredentials::userItem() );
        credentials->second = authenticationService->credentials( connectionString ).valueForItem( IAuthenticationCredentials::passwordItem());
      } else {
        credentials->first = authenticationService->credentials( connectionString, role ).valueForItem( IAuthenticationCredentials::userItem() );
        credentials->second = authenticationService->credentials( connectionString, role ).valueForItem( IAuthenticationCredentials::passwordItem()  );
      }
    } catch (const coral::AuthenticationServiceException& exc) {
      delete credentials;
      credentials = 0;
      throw;
    }
  } else {
    delete credentials;
    credentials = 0;
    log << seal::Msg::Error << "No authentication service has been loaded." << seal::flush;
    throw ConnectionServiceException("No authentication service has been loaded.","ReplicaCatalogue::userCredentials");
  }
  return credentials;
}



/// returns an array of connection strings - the replicas of the specified connection
std::vector<coral::ConnectionService::ConnectionParams>*
coral::ConnectionService::ReplicaCatalogue::getReplicas( const std::string& connectionString,
                                                         const std::string& role,
                                                         AccessMode accessMode ) const {
  seal::MessageStream log( m_serviceConfiguration.serviceContext(),
                           m_serviceConfiguration.serviceName(),
                           seal::Msg::Verbose );
  std::vector<ConnectionParams>*  replicas = new std::vector<ConnectionParams>;
  bool readOnlyOnUpdate = m_serviceConfiguration.isReadOnlySessionOnUpdateConnectionsEnabled();
  if(isConnectionPhysical( connectionString )) {
    coral::IRelationalDomain& domain =
      m_serviceConfiguration.relationalService()->domainForConnection( connectionString );
    std::pair<std::string,std::string> connParams = domain.decodeUserConnectionString( connectionString );

    std::string userName("");
    std::string password("");
    bool authenticated = false;
    if(domain.isAuthenticationRequired( connParams.first )){
      std::pair<std::string,std::string>* credentials = getCredentials( connectionString, role );
      if(credentials){
        authenticated = true;
        userName = credentials->first;
        password = credentials->second;
        delete credentials;
      } else {
        // should never happen...
        throw ConnectionServiceException("Could not find authentication credentials.","ReplicaCatalogue::getReplicas");
      }
    } else {
      authenticated = true;
    }
    if(authenticated) replicas->push_back( ConnectionParams(connectionString,
                                                            connParams.first,
                                                            domain.flavorName(),
                                                            connParams.second,
                                                            userName,
                                                            password) );
  } else {
    log << seal::Msg::Debug << "Looking up "<<
      ((accessMode==coral::ReadOnly)? std::string("read-only") : std::string("update") )<<
     " replicas for logical connection: \""<< connectionString<<"\""<< seal::flush;
    seal::IHandle<coral::ILookupService> lookupService = m_serviceConfiguration.lookupService();
    if( lookupService ){
      
      coral::IDatabaseServiceSet* dbSet = lookupService->lookup( connectionString,
                                                                 accessMode );
      int numOfReplicas = dbSet->numberOfReplicas();
      std::vector<const coral::IDatabaseServiceDescription*> sortedReplicaSet;
      for(int j=0;j<numOfReplicas;j++){
        sortedReplicaSet.push_back(&dbSet->replica( j ));
      }
      coral::IReplicaSortingAlgorithm* sortingAlgorithm = m_serviceConfiguration.sortingAlgorithm();
      if(sortingAlgorithm) {
        log << seal::Msg::Debug << "Re-ordering replicas using installed algorithm."<< seal::flush;
        sortingAlgorithm->sort(sortedReplicaSet);
      }
      numOfReplicas = sortedReplicaSet.size();
      if( numOfReplicas>0 ){
        // set the number of replicas to 1 in update mode or if failover is not enabled
        if((accessMode == coral::Update) || !m_serviceConfiguration.isReplicaFailoverEnabled()) numOfReplicas = 1;
      }
      for(int i=0;i<numOfReplicas;i++){
        std::string replicaConnectionString = sortedReplicaSet[ i ]->connectionString();
        if(replicaConnectionString.empty()) continue;
        coral::IRelationalDomain& domain =
          m_serviceConfiguration.relationalService()->domainForConnection( replicaConnectionString );
        std::pair<std::string,std::string> connParams = domain.decodeUserConnectionString( replicaConnectionString );

        if((accessMode==coral::Update) || // update capable connection are selected by the lookup
           readOnlyOnUpdate || // for read-only connection the lookup will also consider the update
           (dbSet->replica( i ).accessMode()==coral::ReadOnly)) // otherwise strictly select the read-only capable connection
        {
          std::string userName("");
          std::string password("");
          bool authenticated = false;

          if(domain.isAuthenticationRequired( connParams.first )){
            try {
              std::pair<std::string,std::string>* credentials = getCredentials( replicaConnectionString, role );
              if(credentials){
                authenticated = true;
                userName = credentials->first;
                password = credentials->second;
                delete credentials;
              }
            } catch (const AuthenticationServiceException& exc){
              log << seal::Msg::Debug << "Could not retrieve credentials for connection \""<<replicaConnectionString<<"\" from authentication service: "<< exc.what() << seal::flush;
            }
          } else {
            authenticated = true;
          }
          if(authenticated) {
            ConnectionParams connPars(replicaConnectionString,
                                      connParams.first,
                                      domain.flavorName(),
                                      connParams.second,
                                      userName,password);
            // loading service-specific parameters
            std::string connRetrPeriod =
              sortedReplicaSet[ i ]->serviceParameter(coral::IDatabaseServiceDescription::connectionRetrialPeriodParam());
            if(!connRetrPeriod.empty())
              connPars.serviceParameters().insert(
                std::make_pair(coral::IDatabaseServiceDescription::connectionRetrialPeriodParam(),connRetrPeriod));
            std::string connRetrTime =
              sortedReplicaSet[ i ]->serviceParameter(coral::IDatabaseServiceDescription::connectionRetrialTimeOutParam());
            if(!connRetrTime.empty())
              connPars.serviceParameters().insert(
                std::make_pair(coral::IDatabaseServiceDescription::connectionRetrialTimeOutParam(),connRetrTime));
            std::string connTime =
              sortedReplicaSet[ i ]->serviceParameter(coral::IDatabaseServiceDescription::connectionTimeOutParam());
            if(!connTime.empty())
              connPars.serviceParameters().insert(
                std::make_pair(coral::IDatabaseServiceDescription::connectionTimeOutParam(),connTime));
            replicas->push_back( connPars );
          } else {
            log << seal::Msg::Warning << "Replica on connection \""<<replicaConnectionString<<"\" skipped (no authentication provided)." << seal::flush;
          }
        }
      }
      delete dbSet;
      if( replicas->empty() ){
        throw coral::ReplicaNotAvailableException( connectionString ,accessMode,"ReplicaCatalogue::replicasForConnection");
      }
    } else {
      throw coral::ConnectionServiceException("No Lookupservice available.",
                                              "ReplicaCatalogue::replicasForConnection");      
    }
  }
  return replicas;
}

