#include "SessionProxy.h"
#include "ConnectionService.h"
#include "InvalidSessionProxyException.h"
#include "ConnectionPool.h"
#include "RelationalAccess/ISession.h"

/// constructor
coral::ConnectionService::SessionProperties::SessionProperties(const SessionHandle& sessionHandle):
  m_session(sessionHandle){}

/// destructor
coral::ConnectionService::SessionProperties::~SessionProperties(){}

/// Returns the name of the RDBMS flavour..
std::string coral::ConnectionService::SessionProperties::flavorName() const {
  return m_session.connection().technologyName();
}

/// Returns the version of the database server.
std::string coral::ConnectionService::SessionProperties::serverVersion() const {
  return m_session.connection().serverVersion();
}

coral::ConnectionService::SessionProxy::SessionProxy( const std::string& connectionString,
                                                      coral::AccessMode accessMode,
                                                      ConnectionService* connectionService ):
  m_connectionString( connectionString ),
  m_role(""),
  m_accessMode( accessMode ),
  m_connectionService( connectionService ),
  m_session(),
  m_properties( m_session )
{
}

coral::ConnectionService::SessionProxy::SessionProxy( const std::string& connectionString,
                                                      const std::string& role,
                                                      coral::AccessMode accessMode,
                                                      ConnectionService* connectionService ):
  m_connectionString( connectionString ),
  m_role( role ),
  m_accessMode( accessMode ),
  m_connectionService( connectionService ),
  m_session(),
  m_properties( m_session )
{
}

coral::ConnectionService::SessionProxy::~SessionProxy()
{
  m_session.close();
  if(m_connectionService) {
    m_connectionService->connectionPool().releaseConnection( m_session.connection() );
    m_connectionService->unRegisterSession( this );
  }
}

void coral::ConnectionService::SessionProxy::open(){
  if( ! m_connectionService ) throw InvalidSessionProxyException(m_connectionString,"SessionProxy::properties" );
  m_session = m_connectionService->connectionPool().getValidSession( m_connectionString, m_role, m_accessMode );  
}

/// Returns the session info object.
coral::ISessionProperties& coral::ConnectionService::SessionProxy::properties(){
  if (!m_session.isValid())
  {
    if( ! m_connectionService ) {
      throw InvalidSessionProxyException(m_connectionString,"SessionProxy::properties" );
    }
    m_connectionService->connectionPool().removeConnection( m_session.connection() );
    m_session = m_connectionService->connectionPool().getValidSession( m_connectionString, m_role, m_accessMode );    
  }
  return m_properties;
}

coral::ISchema& coral::ConnectionService::SessionProxy::nominalSchema()
{
  if ( !m_session.isValid() )
  {
    if( ! m_connectionService ) {
      throw InvalidSessionProxyException(m_connectionString,"SessionProxy::nominalSchema" );
    }
    m_connectionService->connectionPool().removeConnection( m_session.connection() );
    m_session = m_connectionService->connectionPool().getValidSession( m_connectionString, m_role, m_accessMode );    
  }
  return m_session.physicalSession()->nominalSchema();
}

coral::ISchema& coral::ConnectionService::SessionProxy::schema(const std::string& schemaName )
{
  if (!m_session.isValid())
  {
    if( ! m_connectionService ) {
      throw InvalidSessionProxyException(m_connectionString,"SessionProxy::schema" );
    }
    m_connectionService->connectionPool().removeConnection( m_session.connection() );
    m_session = m_connectionService->connectionPool().getValidSession( m_connectionString, m_role, m_accessMode );    
  }
  return m_session.physicalSession()->schema( schemaName );
}
 
/** Check if connection is shared between users */ 
bool coral::ConnectionService::SessionProxy::isConnectionShared() const
{
  if(!m_session.isValid()) {
    throw InvalidSessionProxyException(m_connectionString,"SessionProxy::isConnectionShared" );
  }
  return m_session.connection().numberOfSessions()>1;
}

coral::ITransaction&  coral::ConnectionService::SessionProxy::transaction()
{
  if (!m_session.isValid())
  {
    if( ! m_connectionService ) {
      throw InvalidSessionProxyException(m_connectionString,"SessionProxy::transaction" );
    }
    m_connectionService->connectionPool().removeConnection( m_session.connection() );
    m_session = m_connectionService->connectionPool().getValidSession( m_connectionString, m_role, m_accessMode );    
  }
  return m_session.physicalSession()->transaction();
}

coral::ITypeConverter& coral::ConnectionService::SessionProxy::typeConverter()
{
  if (!m_session.isValid())
  {
    if( ! m_connectionService ) {
      throw InvalidSessionProxyException(m_connectionString,"SessionProxy::typeConverter" );
    }
    m_connectionService->connectionPool().removeConnection( m_session.connection() );
    m_session = m_connectionService->connectionPool().getValidSession( m_connectionString, m_role, m_accessMode );    
  }
  return m_session.connection().typeConverter();
}

/// hook to access the underline session object
coral::ISession* coral::ConnectionService::SessionProxy::session(){
  return m_session.physicalSession();
}

/// invalidate the stored handle to the Connection service
void coral::ConnectionService::SessionProxy::invalidate(){
  m_connectionService = 0;
}

