// -*- C++ -*-
// $Id: Connection.cpp,v 1.5 2007/02/13 16:37:04 rado Exp $
#include "Connection.h"
#include "Session.h"
#include "DomainProperties.h"
#include "ErrorHandler.h"
#include "TypeConverter.h"

#include "CoralCommon/URIParser.h"
#include "CoralCommon/MonitoringEventDescription.h"

#include "RelationalAccess/SessionException.h"
#include "RelationalAccess/IWebCacheControl.h"
#include "RelationalAccess/IWebCacheInfo.h"
#include "RelationalAccess/AccessMode.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"

#include "SealKernel/Service.h"
#include "SealKernel/Context.h"
#include "SealKernel/MessageStream.h"

#include "frontier_client/frontier-cpp.h"
#include "frontier_client/FrontierException.hpp"

#include <locale>

coral::FrontierAccess::Connection::Connection( const coral::FrontierAccess::DomainProperties& domainProperties, const std::string& connectionString )
  : m_connection(0)
  , m_domainProperties( domainProperties )
  , m_connectionString( connectionString )
  , m_connected( false )
  , m_serverVersion( "" )
  , m_typeConverter( new coral::FrontierAccess::TypeConverter( m_domainProperties ) )
  , m_lock()
{
  if( this->m_typeConverter )
    this->m_typeConverter->reset( 10 );
}

coral::FrontierAccess::Connection::~Connection()
{
  this->disconnect();
  delete this->m_typeConverter;
}

void coral::FrontierAccess::Connection::connect()
{
  seal::MessageStream log( m_domainProperties.service(), m_domainProperties.service()->name(), seal::Msg::Verbose );  

  // Locate the web cache control and set up the compression level
  try
  {
    coral::IWebCacheControl& cachectrl = const_cast<coral::IWebCacheControl&>( m_domainProperties.cacheControl() );
    const char* zipLevelStr = 0;
    zipLevelStr = ::getenv( "CORAL_WEBCACHE_ZIP_LEVEL" );
    if( zipLevelStr != 0 )
    {
      int level = 5;
      level = ::atoi( zipLevelStr );
      log << seal::Msg::Warning << "Changing the compression level from " << cachectrl.compressionLevel() << " to level " << level << seal::flush;
      cachectrl.setCompressionLevel( level );
    }
    else
    {
      log << seal::Msg::Info << "The current compresion level is " << cachectrl.compressionLevel() << seal::flush;
    }
  }
  catch( const std::exception& e )
  {
    seal::MessageStream log( m_domainProperties.service(), m_domainProperties.service()->name(), seal::Msg::Debug );
    log << seal::Msg::Warning
        << "The web cache control cannot be located, check if the coral::ConnectionService is loaded or web cache control set up, running with the default compression level"
        << seal::flush;
  }  

  std::string furi = "frontier://" + m_connectionString;
  
  // Processing the connection string
  coral::URIParser parser; parser.setURI( furi );
  
  std::ostringstream osDatabase;
  osDatabase << parser.hostName();
  int portNumber = parser.portNumber();
  if ( portNumber != 0 ) osDatabase << ":" << portNumber;

  std::string constr = furi;
//   size_t      pos    = constr.rfind( "/" );

//   if( pos == std::string::npos )
//     throw coral::DatabaseNotAccessibleException( m_domainProperties.service()->name(),
//                                                  "Invalid connection string: " + constr + ", can't extract schema name", "coral::FrontierAccess::Connection::connect" );
  
//   std::string schema = constr.substr( pos+1 );

//   if( schema.empty() )
//     throw coral::DatabaseNotAccessibleException( m_domainProperties.service()->name(),
//                                                  "Invalid connection string: " + constr + ", no schema name specified", "coral::FrontierAccess::Connection::connect" );

//   std::string conurl = "http" + constr.substr( 8, constr.size() - schema.size() - 8 - 1 );
  
//   if( conurl.empty() )
//     throw coral::DatabaseNotAccessibleException( m_domainProperties.service()->name(),
//                                                  "Invalid connection string: " + constr + ", can't translate into correct Frontier URL", "coral::FrontierAccess::Connection::connect" );
  
//   log << seal::Msg::Verbose << "Connecting to Frontier server using URL: " << conurl << seal::flush;

  log << seal::Msg::Verbose << "Connecting to Frontier server using URL: " << this->m_connectionString << seal::flush;
  
  {
    boost::mutex::scoped_lock   lock( m_lock );
    
    // Attaching the server
    m_connection = new frontier::Connection( m_connectionString );
    
    // With Frontier does not matter
    m_serverVersion="10.2.0.2.0";
    
    this->m_connected = true;
  }
}

coral::ISession* coral::FrontierAccess::Connection::newSession( const std::string& schemaName, coral::AccessMode mode/*= coral::Update*/ ) const
{
  coral::FrontierAccess::Session*
    session = new coral::FrontierAccess::Session( const_cast<coral::FrontierAccess::Connection&>(*this),
                                                  m_domainProperties,
                                                  m_connectionString,
                                                  *m_connection,
                                                  m_lock,
                                                  schemaName,
                                                  *m_typeConverter,
                                                  mode );
  return session;
}

bool coral::FrontierAccess::Connection::isConnected( bool probePhysicalConnection )
{  
  if( probePhysicalConnection )
  {
    seal::MessageStream log( m_domainProperties.service(), "coral::FrontierAccess::Connection::isConnected", seal::Msg::Verbose );

    try
    {
      this->m_connected = this->ping();
    }
    catch( const frontier::FrontierException& fe )
    {
      m_connected = false;
      const char* errmsg = fe.what();
      if( errmsg != 0 )
        log << seal::Msg::Error << fe.what() << seal::flush;
      else
        log << seal::Msg::Error << "coral::FrontierAccess::Connection: Error checking connection" << seal::flush;
    }
  }

  return m_connected;
}

void coral::FrontierAccess::Connection::disconnect()
{
  if( ! m_connected )
    return;

  if( m_connection )
  {
    delete m_connection;
    m_connection = 0;
  }
  
  m_connected = false;    
}

std::string coral::FrontierAccess::Connection::serverVersion() const
{
  if ( ! const_cast<coral::FrontierAccess::Connection*>( this )->isConnected() )
    throw coral::ConnectionNotActiveException( m_domainProperties.service()->name(), "coral::FrontierAccess::Connection::serverVersion" );
  return m_serverVersion;
}

coral::ITypeConverter& coral::FrontierAccess::Connection::typeConverter()
{
  if ( ! this->isConnected() )
    throw coral::ConnectionNotActiveException( m_domainProperties.service()->name(), "coral::FrontierAccess::Connection::typeConverter" );
  return *(this->m_typeConverter);
}

bool coral::FrontierAccess::Connection::ping()
{
  bool available = true;

//   // Use a statement to find the Oracle server version behind the Frontier app. server
//   coral::AttributeList inBuffer;
//   coral::FrontierAccess::Statement query( *m_properties, "SELECT 1 FROM DUAL" ); // FIXME !!!!
//   this->m_connected = query.execute( inBuffer, false /* rely on cache */ ); // We realy now on the Frontier client + caching to avoid Oracle server acess overhead
  
  return available;
}
