#include "Transaction.h"
#include "SessionProperties.h"
#include "DomainProperties.h"
#include "ErrorHandler.h"
#include "ITransactionObserver.h"

#include "CoralCommon/MonitoringEventDescription.h"

#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/SessionException.h"

#include "SealKernel/MessageStream.h"
#include "SealKernel/Service.h"
#include "SealBase/TimeInfo.h"

coral::FrontierAccess::Transaction::Transaction( coral::FrontierAccess::SessionProperties& properties, ITransactionObserver& observer )
  : m_sessionProperties( properties ), m_observer( observer ), m_isReadOnly( false ), m_isActive( false )
{
}

coral::FrontierAccess::Transaction::~Transaction()
{
  if ( this->isActive() )
    this->rollback();
}

void coral::FrontierAccess::Transaction::start( bool readOnly )
{
  seal::MessageStream log( m_sessionProperties.domainProperties().service(), m_sessionProperties.domainProperties().service()->name() );

  if ( this->isActive() )
  {
    log << seal::Msg::Warning << "A transaction is already active" << seal::flush;
    return;
  }

  m_isReadOnly = readOnly;

  if ( m_sessionProperties.monitoringService() )
  {
    m_sessionProperties.monitoringService()->record( m_sessionProperties.connectionString(), coral::monitor::Transaction, coral::monitor::Info,
                                                     monitoringEventDescription.sessionBegin(), ( m_isReadOnly ? monitoringEventDescription.transactionReadOnly() : monitoringEventDescription.transactionUpdate() ) );
  }

  this->m_isActive = true;
}

void coral::FrontierAccess::Transaction::commit()
{
  seal::MessageStream log( m_sessionProperties.domainProperties().service(), m_sessionProperties.domainProperties().service()->name() );

  if ( ! ( this->isActive() ) )
  {
    log << seal::Msg::Warning << "No active transaction to commit" << seal::flush;
    return;
  }

  seal::TimeInfo::NanoSecs t0 = seal::TimeInfo::processIdleTime();
  // Do nothing
  ;;;
  seal::TimeInfo::NanoSecs t1 = seal::TimeInfo::processIdleTime();
  double idleTimeSeconds = ( t1-t0 ) * 1e-9;
  if ( m_sessionProperties.monitoringService() )
  {
    m_sessionProperties.monitoringService()->record( m_sessionProperties.connectionString(), coral::monitor::Transaction, coral::monitor::Time, monitoringEventDescription.transactionCommit(), idleTimeSeconds );
  }

  m_observer.reactOnEndOfTransaction();

  this->m_isActive = false;
}

void coral::FrontierAccess::Transaction::rollback()
{
  seal::MessageStream log( m_sessionProperties.domainProperties().service(), m_sessionProperties.domainProperties().service()->name() );

  if ( ! ( this->isActive() ) )
  {
    log << seal::Msg::Warning << "No active transaction to roll back" << seal::flush;
    return;
  }

  if ( m_sessionProperties.monitoringService() )
  { 
    m_sessionProperties.monitoringService()->record( m_sessionProperties.connectionString(), coral::monitor::Transaction, coral::monitor::Info, monitoringEventDescription.transactionRollback() );
  }

  m_observer.reactOnEndOfTransaction();

  this->m_isActive = false;
}

void coral::FrontierAccess::Transaction::allocateHandles() { ; }

void coral::FrontierAccess::Transaction::releaseHandles() { ; }

bool coral::FrontierAccess::Transaction::isActive() const { return m_isActive; }

bool coral::FrontierAccess::Transaction::isReadOnly() const { return m_isReadOnly; }
