#include "Connection.h"
#include "Session.h"
#include "ConnectionProperties.h"
#include "DomainProperties.h"
#include "OracleErrorHandler.h"
#include "CoralCommon/MonitoringEventDescription.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/SessionException.h"
#include "SealKernel/Service.h"
#include "SealKernel/Context.h"
#include "SealKernel/MessageStream.h"
#include "oci.h"
#include <sstream>

coral::OracleAccess::Connection::Connection( const coral::OracleAccess::DomainProperties& domainProperties,
                                             const std::string& connectionString ):
  m_properties( new coral::OracleAccess::ConnectionProperties( domainProperties,
                                                               connectionString ) ),
  m_serverVersion( "" ),
  m_mutex()
{}


coral::OracleAccess::Connection::~Connection()
{
  if ( this->isConnected() )
    this->disconnect();
  delete m_properties;
}


coral::ISession*
coral::OracleAccess::Connection::newSession( const std::string& schemaName,
                                             coral::AccessMode mode ) const
{
  if ( ! const_cast<coral::OracleAccess::Connection*>( this )->isConnected() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(),
                                               "IConnection::newSession" );

  boost::mutex::scoped_lock lock( m_mutex );
  return new coral::OracleAccess::Session( *( static_cast<coral::IDevConnection*>( const_cast<coral::OracleAccess::Connection*>( this ) ) ),
                                           *m_properties,
                                           schemaName,
                                           mode );
}


void
coral::OracleAccess::Connection::connect()
{
  if ( this->isConnected() ) {
    seal::MessageStream log( m_properties->domainProperties().service(),
                             m_properties->domainProperties().service()->name() );
    log << seal::Msg::Warning << "A connection is already present for \""
        << m_properties->connectionString() << "\"" << seal::flush;
    return;
  }

  boost::mutex::scoped_lock lock( m_mutex );

  const std::string database = m_properties->connectionString();

  // Creating the environment
  OCIEnv* environmentHandle = 0;
  sword status = OCIEnvCreate( &environmentHandle,
                               OCI_OBJECT | OCI_THREADED,
                               0,0,0,0,0,0 );
  if ( status != OCI_SUCCESS ) {
    throw coral::ServerException( m_properties->domainProperties().service()->name(),
                                  "Could not allocate an OCI environment handle" );
  }

  // Creating the error handle
  void* temporaryPointer = 0;
  status = OCIHandleAlloc( environmentHandle, &temporaryPointer,
                           OCI_HTYPE_ERROR, 0, 0 );
  if ( status != OCI_SUCCESS ) {
    OCIHandleFree( environmentHandle, OCI_HTYPE_ENV );
    throw coral::ServerException( m_properties->domainProperties().service()->name(),
                                  "Could not allocate an OCI error handle" );
  }
  OCIError* errorHandle = static_cast< OCIError* >( temporaryPointer );

  // Creating a server handle
  temporaryPointer = 0;
  status = OCIHandleAlloc( environmentHandle, &temporaryPointer,
                           OCI_HTYPE_SERVER, 0, 0 );
  if ( status != OCI_SUCCESS ) {
    OCIHandleFree( errorHandle, OCI_HTYPE_ERROR );
    OCIHandleFree( environmentHandle, OCI_HTYPE_ENV );
    throw coral::ServerException( m_properties->domainProperties().service()->name(),
                                  "Could not allocate an OCI server handle" );
  }
  
  OCIServer* serverHandle = static_cast< OCIServer* >( temporaryPointer );
  
  // Creating an error handler to be used in the subsequent calls.
  coral::OracleAccess::OracleErrorHandler errorHandler( errorHandle );

  // Attaching the server
  status = OCIServerAttach( serverHandle, errorHandle,
                            reinterpret_cast<const text *>( database.c_str() ), ::strlen( database.c_str() ),
                            OCI_DEFAULT );
  if ( status != OCI_SUCCESS ) {
    errorHandler.handleCase( status, "attaching a server" );
    seal::MessageStream log( m_properties->domainProperties().service(),
                             m_properties->domainProperties().service()->name() );
    if ( errorHandler.isError() ) {
      log << seal::Msg::Error << errorHandler.message() << seal::flush;
      OCIHandleFree( serverHandle, OCI_HTYPE_SERVER );
      OCIHandleFree( errorHandle, OCI_HTYPE_ERROR );
      OCIHandleFree( environmentHandle, OCI_HTYPE_ENV );
      const sb4 lastErrorCode = errorHandler.lastErrorCode();
      if ( lastErrorCode == 12154 || lastErrorCode == 12538 ||
           lastErrorCode == 12545 || lastErrorCode == 12505) { // Wrong db specification
        throw coral::DatabaseNotAccessibleException( m_properties->domainProperties().service()->name(),
                                                     "IConnection::connect" );
      }
      if ( true ) { // Retry in these cases
        throw coral::ConnectionException( m_properties->domainProperties().service()->name(),
                                          "IConnection::connect" );
      }
      else { // Do not retry
        throw coral::ServerException( m_properties->domainProperties().service()->name() );
      }
    }
    else {
      log << seal::Msg::Warning << errorHandler.message() << seal::flush;
    }
  }

  // Retrieving the server version
  text serverVersion[1000];
  status = OCIServerVersion( serverHandle, errorHandle, serverVersion, 1000, OCI_HTYPE_SERVER );
  if ( status != OCI_SUCCESS ) {
    seal::MessageStream log( m_properties->domainProperties().service(),
                             m_properties->domainProperties().service()->name() );
    log << seal::Msg::Error << "Failed to retrieve the server version" << seal::flush;
    OCIServerDetach( serverHandle, errorHandle, OCI_DEFAULT );
    OCIHandleFree( serverHandle, OCI_HTYPE_SERVER );
    OCIHandleFree( errorHandle, OCI_HTYPE_ERROR );
    OCIHandleFree( environmentHandle, OCI_HTYPE_ENV );
    throw coral::ConnectionException( m_properties->domainProperties().service()->name(),
                                      "ISession::connect",
                                      "Could not retrieve the Oracle server version" );
  }
  std::ostringstream os;
  os << serverVersion;
  int majorVersionNumber = this->parseVersionString( os.str() );

  m_properties->setHandles( environmentHandle, errorHandle, serverHandle, majorVersionNumber );

  // Record the beginning of the connection using the Monitoring service
  std::vector< seal::IHandle< coral::monitor::IMonitoringService > > loadedMonitoringServices;
  m_properties->domainProperties().service()->context()->query( loadedMonitoringServices );
  if ( ! loadedMonitoringServices.empty() ) {
    coral::monitor::IMonitoringService& monitoringService = *( loadedMonitoringServices.front().get() );
    monitoringService.setLevel( "oracle://" + m_properties->connectionString(), coral::monitor::Default );
    monitoringService.enable( "oracle://" + m_properties->connectionString() );
    monitoringService.record( "oracle://" + m_properties->connectionString(),
                              coral::monitor::Session,
                              coral::monitor::Info,
                              monitoringEventDescription.connectionBegin() );
  }
}

bool
coral::OracleAccess::Connection::isConnected( bool probePhysicalConnection )
{
  boost::mutex::scoped_lock lock( m_mutex );

  bool logicalConnection = ( m_properties->environmentHandle() != 0 );
  if ( ! logicalConnection ) return false;
  if ( probePhysicalConnection ) {
    text serverVersion[1000];
    sword status = OCIServerVersion( m_properties->serverHandle(), m_properties->errorHandle(),
                                     serverVersion, 1000, OCI_HTYPE_SERVER );
    if ( status != OCI_SUCCESS ) {
      this->__disconnect( true );
      logicalConnection = false;
    }
  }

  return logicalConnection;
}


void
coral::OracleAccess::Connection::disconnect()
{
  if ( this->isConnected() ) {
    boost::mutex::scoped_lock lock( m_mutex );
    this->__disconnect();
  }
}


void
coral::OracleAccess::Connection::__disconnect( bool lostConnection )
{
  // Clean up all the active sessions !!!
  this->invalidateAll();
  
  // Close all handles.
  OCIEnv* environmentHandle = m_properties->environmentHandle();
  OCIError* errorHandle = m_properties->errorHandle();
  OCIServer* serverHandle = m_properties->serverHandle();
  OCIServerDetach( serverHandle, errorHandle, OCI_DEFAULT );
  if ( ! lostConnection ) { // THIS IS NECESSARY OTHERWISE OCI WILL DOUBLE FREE THE HANDLE!!!
    OCIHandleFree( serverHandle, OCI_HTYPE_SERVER );
  }
  OCIHandleFree( errorHandle, OCI_HTYPE_ERROR );
  OCIHandleFree( environmentHandle, OCI_HTYPE_ENV );
  
  // Record the ending of the connection
  std::vector< seal::IHandle< coral::monitor::IMonitoringService > > loadedMonitoringServices;
  m_properties->domainProperties().service()->context()->query( loadedMonitoringServices );
  if ( ! loadedMonitoringServices.empty() ) {
    coral::monitor::IMonitoringService& monitoringService = *( loadedMonitoringServices.front().get() );
    monitoringService.record( "oracle://" + m_properties->connectionString(),
                              coral::monitor::Session,
                              coral::monitor::Info,
                              monitoringEventDescription.connectionEnd() );
    monitoringService.disable( "oracle://" + m_properties->connectionString() );
  }

  m_properties->setHandles( 0, 0, 0, false );
}


std::string
coral::OracleAccess::Connection::serverVersion() const
{
  if ( ! const_cast<coral::OracleAccess::Connection*>( this )->isConnected() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(),
                                               "IConnection::serverVersion" );
  boost::mutex::scoped_lock lock( m_mutex );
  return m_serverVersion;
}


coral::ITypeConverter&
coral::OracleAccess::Connection::typeConverter()
{
  if ( ! this->isConnected() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(),
                                               "Connection::typeConverter" );
  boost::mutex::scoped_lock lock( m_mutex );
  return m_properties->typeConverter();
}


int
coral::OracleAccess::Connection::parseVersionString( const std::string& versionString )
{
  // Trying to find a number of the format "i1.i2.i3.i4.i5"
  std::istringstream is( versionString.c_str() );
  while ( ! is.eof() ) {
    std::string result;
    is >> result >> std::ws;
    std::istringstream isnumbers( result.c_str() );
    int i1, i2, i3, i4, i5;
    char c = ' ';
    isnumbers >> i1 >> c;
    if ( c != '.' ) continue;
    c = ' ';
    isnumbers >> i2 >> c;
    if ( c != '.' ) continue;
    c = ' ';
    isnumbers >> i3 >> c;
    if ( c != '.' ) continue;
    c = ' ';
    isnumbers >> i4 >> c;
    if ( c != '.' ) continue;
    isnumbers >> i5;
    std::ostringstream os;
    os << i1 << c << i2 << c << i3 << c << i4 << c << i5;
    m_serverVersion = os.str();
    return i1;
  }
  return 0;
}
