#include "Transaction.h"
#include "SessionProperties.h"
#include "DomainProperties.h"
#include "OracleErrorHandler.h"
#include "ITransactionObserver.h"
#include "CoralCommon/MonitoringEventDescription.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/SessionException.h"
#include "SealKernel/MessageStream.h"
#include "SealKernel/Service.h"
#include "SealBase/TimeInfo.h"
#include "oci.h"


coral::OracleAccess::Transaction::Transaction( coral::OracleAccess::SessionProperties& properties,
                                               ITransactionObserver& observer ):
  m_sessionProperties( properties ),
  m_observer( observer ),
  m_transactionHandle( 0 ),
  m_isReadOnly( false ),
  m_mutex()
{}


coral::OracleAccess::Transaction::~Transaction()
{
  if ( this->isActive() ) this->rollback();
}


void
coral::OracleAccess::Transaction::start( bool readOnly )
{
  if ( this->isActive() ) {
    seal::MessageStream log( m_sessionProperties.domainProperties().service(),
                             m_sessionProperties.domainProperties().service()->name() );
    log << seal::Msg::Warning << "A transaction is already active" << seal::flush;
    return;
  }

  boost::mutex::scoped_lock lock(m_mutex);

  if ( ! readOnly && m_sessionProperties.isReadOnly() ) {
    throw coral::InvalidOperationInReadOnlyModeException( m_sessionProperties.domainProperties().service()->name(),
                                                          "ITransaction::start" );
  }

  this->allocateHandles();
  sword status = OCITransStart( m_sessionProperties.serviceContextHandle(),
                                m_sessionProperties.errorHandle(),
                                0,
                                OCI_TRANS_NEW | ( readOnly ? OCI_TRANS_READONLY : 0 ) );
  if ( status != OCI_SUCCESS ) {
    seal::MessageStream log( m_sessionProperties.domainProperties().service(),
                             m_sessionProperties.domainProperties().service()->name() );
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "start transaction" );
    if ( errorHandler.isError() ) {
      log << seal::Msg::Error << errorHandler.message() << seal::flush;
      this->releaseHandles();
      throw coral::TransactionNotStartedException( m_sessionProperties.domainProperties().service()->name() );
    }
    log << seal::Msg::Warning << errorHandler.message() << seal::flush;
  }

  m_isReadOnly = readOnly;
  if ( m_sessionProperties.monitoringService() ) {
    m_sessionProperties.monitoringService()->record( "oracle://" + m_sessionProperties.connectionString() + "/" + m_sessionProperties.schemaName(),
                                                     coral::monitor::Transaction,
                                                     coral::monitor::Info,
                                                     monitoringEventDescription.sessionBegin(),
                                                     ( m_isReadOnly ? monitoringEventDescription.transactionReadOnly() : monitoringEventDescription.transactionUpdate() ) );
  }

}


void
coral::OracleAccess::Transaction::commit()
{
  if ( ! ( this->isActive() ) ) {
    seal::MessageStream log( m_sessionProperties.domainProperties().service(),
                             m_sessionProperties.domainProperties().service()->name() );
    log << seal::Msg::Warning << "No active transaction to commit" << seal::flush;
    return;
  }

  boost::mutex::scoped_lock lock(m_mutex);
  seal::TimeInfo::NanoSecs t0 = 0;
  if ( m_sessionProperties.monitoringService() ) t0 = seal::TimeInfo::processIdleTime();
  sword status = OCITransCommit( m_sessionProperties.serviceContextHandle(),
                                 m_sessionProperties.errorHandle(),
                                 OCI_DEFAULT );
  if ( m_sessionProperties.monitoringService() ) {
    seal::TimeInfo::NanoSecs t1 = seal::TimeInfo::processIdleTime();
    double idleTimeSeconds = ( t1-t0 ) * 1e-9;
    m_sessionProperties.monitoringService()->record( "oracle://" + m_sessionProperties.connectionString() + "/" + m_sessionProperties.schemaName(),
                                                     coral::monitor::Transaction,
                                                     coral::monitor::Time,
                                                     monitoringEventDescription.transactionCommit(),
                                                     idleTimeSeconds );
  }

  m_observer.reactOnEndOfTransaction();

  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "commit transaction" );
    seal::MessageStream log( m_sessionProperties.domainProperties().service(),
                             m_sessionProperties.domainProperties().service()->name() );
    log << seal::Msg::Error << errorHandler.message() << seal::flush;
    this->releaseHandles();
    throw coral::TransactionNotCommittedException( m_sessionProperties.domainProperties().service()->name() );
  }

  this->releaseHandles();
}


void
coral::OracleAccess::Transaction::rollback()
{
  if ( ! ( this->isActive() ) ) {
    seal::MessageStream log( m_sessionProperties.domainProperties().service(),
                             m_sessionProperties.domainProperties().service()->name() );
    log << seal::Msg::Warning << "No active transaction to roll back" << seal::flush;
    return;
  }

  boost::mutex::scoped_lock lock(m_mutex);
  sword status = OCITransRollback( m_sessionProperties.serviceContextHandle(),
                                   m_sessionProperties.errorHandle(),
                                   OCI_DEFAULT );
  if ( m_sessionProperties.monitoringService() ) {
    m_sessionProperties.monitoringService()->record( "oracle://" + m_sessionProperties.connectionString() + "/" + m_sessionProperties.schemaName(),
                                                     coral::monitor::Transaction,
                                                     coral::monitor::Info,
                                                     monitoringEventDescription.transactionRollback() );
  }

  m_observer.reactOnEndOfTransaction();

  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler( m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "roll back transaction" );
    seal::MessageStream log( m_sessionProperties.domainProperties().service(),
                             m_sessionProperties.domainProperties().service()->name() );
    log << seal::Msg::Error << errorHandler.message() << seal::flush;
  }

  this->releaseHandles();
}


void
coral::OracleAccess::Transaction::allocateHandles()
{
  // Allocating a new transaction handle
  void* temporaryPointer = 0;
  sword status = OCIHandleAlloc( m_sessionProperties.environmentHandle(),
                                 &temporaryPointer,
                                 OCI_HTYPE_TRANS, 0, 0 );
  if ( status != OCI_SUCCESS ) {
    throw coral::TransactionNotStartedException( m_sessionProperties.domainProperties().service()->name(),
                                                 "Failed to allocate a new transaction handle" );
  }
  m_transactionHandle = static_cast< OCITrans* >( temporaryPointer );

  // Setting the transaction handle to the service context
  status = OCIAttrSet( m_sessionProperties.serviceContextHandle(), OCI_HTYPE_SVCCTX, 
                       m_transactionHandle, 0, OCI_ATTR_TRANS, m_sessionProperties.errorHandle() );
  if ( status != OCI_SUCCESS ) {
    coral::OracleAccess::OracleErrorHandler errorHandler(  m_sessionProperties.errorHandle() );
    errorHandler.handleCase( status, "Setting the transaction handle to the service context" );
    seal::MessageStream log( m_sessionProperties.domainProperties().service(),
                             m_sessionProperties.domainProperties().service()->name() );
    if ( errorHandler.isError() ) {
      log << seal::Msg::Error << errorHandler.message() << seal::flush;
      OCIHandleFree( m_transactionHandle, OCI_HTYPE_TRANS );
      m_transactionHandle = 0;
      throw coral::TransactionNotStartedException( m_sessionProperties.domainProperties().service()->name(),
                                                   "Failed to set the transaction handle to the service context handle" );
    }
    else {
      log << seal::Msg::Warning << errorHandler.message() << seal::flush;
    }
  }
}


void
coral::OracleAccess::Transaction::releaseHandles()
{
  OCITransDetach( m_sessionProperties.serviceContextHandle(),
                  m_sessionProperties.errorHandle(), OCI_DEFAULT );
  OCIHandleFree( m_transactionHandle, OCI_HTYPE_TRANS );
  m_transactionHandle = 0;
}


bool
coral::OracleAccess::Transaction::isActive() const
{
  boost::mutex::scoped_lock lock(m_mutex);
  return ( m_transactionHandle != 0 );
}
  

bool
coral::OracleAccess::Transaction::isReadOnly() const
{
  boost::mutex::scoped_lock lock(m_mutex);
  return m_isReadOnly;
}
