#include "PyCoral/Context.h"
#include "Exception.h"
#include "SealKernel/Context.h"
#include "SealKernel/ComponentLoader.h"
#include "PluginManager/PluginManager.h"
#include "SealKernel/IMessageService.h"
#include <exception>

// Forward declaration of the methods
static int Context_init( PyObject* self, PyObject* args, PyObject* kwds );
static void Context_dealloc( PyObject* self );
static PyObject* Context_loadComponent( PyObject* self, PyObject* args );
static PyObject* Context_setVerbosityLevel(  PyObject* self, PyObject* args );

// Type definition
PyTypeObject*
coral::PyCoral::Context_Type()
{
  static PyMethodDef Context_Methods[] = {
    { (char*) "loadComponent", (PyCFunction) Context_loadComponent, METH_O,
      (char*) "Loads a component into the context." },
    { (char*) "setVerbosityLevel", (PyCFunction) Context_setVerbosityLevel, METH_O,
      (char*) "Sets the verbosity level in the MessageService. If this service does not exist, it loads it.\nAcceptable input values are: \"VERBOSE\", \"DEBUG\", \"INFO\", \"WARNING\", \"ERROR\", \"FATAL\" and \"ALWAYS\".\nDefault level is \"INFO\"." },
    {0, 0, 0, 0}
  };

  static char Context_doc[] = "A context class.\nIt can be used to load components in the underlying seal::Context";

  static PyTypeObject Context_Type = {
    PyObject_HEAD_INIT(0)
    0,                       /*ob_size*/
    (char*) "coral.Context", /*tp_name*/
    sizeof(coral::PyCoral::Context), /*tp_basicsize*/
    0,                       /*tp_itemsize*/
    /* methods */
    Context_dealloc,         /*tp_dealloc*/
    0,                       /*tp_print*/
    0,                       /*tp_getattr*/
    0,                       /*tp_setattr*/
    0,                       /*tp_compare*/
    0,                       /*tp_repr*/
    0,                       /*tp_as_number*/
    0,                       /*tp_as_sequence*/
    0,                       /*tp_as_mapping*/
    0,                       /*tp_hash*/
    0,                       /*tp_call*/
    0,                       /*tp_str*/
    PyObject_GenericGetAttr, /*tp_getattro*/
    PyObject_GenericSetAttr, /*tp_setattro*/
    0,                       /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT,      /*tp_flags*/
    Context_doc,             /*tp_doc*/
    0,                       /*tp_traverse*/
    0,                       /*tp_clear*/
    0,                       /*tp_richcompare*/
    0,                       /*tp_weaklistoffset*/
    0,                       /*tp_iter*/
    0,                       /*tp_iternext*/
    Context_Methods,         /*tp_methods*/
    0,                       /*tp_members*/
    0,                       /*tp_getset*/
    0,                       /*tp_base*/
    0,                       /*tp_dict*/
    0,                       /*tp_descr_get*/
    0,                       /*tp_descr_set*/
    0,                       /*tp_dictoffset*/
    Context_init,            /*tp_init*/
    PyType_GenericAlloc,     /*tp_alloc*/
    PyType_GenericNew,       /*tp_new*/
    PyObject_Del,           /*tp_free*/
    0,                       /*tp_is_gc*/
    0,                       /*tp_bases*/
    0,                       /*tp_mro*/
    0,                       /*tp_cache*/
    0,                       /*tp_subclasses*/
    0,                       /*tp_weaklist*/
    Context_dealloc          /*tp_del*/
  };
  return &Context_Type;
}

int
Context_init( PyObject* self, PyObject* /* args */, PyObject* /* kwds */ )
{
 coral::PyCoral::Context* py_this = (coral::PyCoral::Context*) self;

 // Initialize here the plugin manager
 seal::PluginManager* pm = seal::PluginManager::get();
 pm->initialise();

 py_this->object = new seal::Context();
 intrusive_ptr_add_ref( py_this->object );
 return 0;
}

void
Context_dealloc( PyObject* self )
{
 coral::PyCoral::Context* py_this = (coral::PyCoral::Context*) self;
 intrusive_ptr_release( py_this->object );
 py_this->object = 0;
 self->ob_type->tp_free( self );
}


PyObject*
Context_loadComponent( PyObject* self, PyObject* args )
{
 coral::PyCoral::Context* py_this = (coral::PyCoral::Context*) self;
 if( ! py_this->object ){
  PyErr_SetString( coral::PyCoral::Exception(),
                   (char*) "Error: No Valid C++ Object" );
  return 0;
 }
 const char* input = 0;
 if (PyString_Check(args))
  input = PyString_AsString(args);
 else{
  PyErr_SetString( coral::PyCoral::Exception(),
                  (char*) "Error: Bad Argument Type" );
  return 0;
 }
 const std::string sinput = std::string(input);
 try {
  seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( py_this->object );
  if ( ! loader->load( sinput ) ) {
   PyErr_SetString( coral::PyCoral::Exception(),
                   ( std::string( "An error occured when loading " ) + sinput ).c_str() );
   return 0;
  }
 }
 catch ( std::exception& e ) {
  PyErr_SetString( coral::PyCoral::Exception(), e.what() );
  return 0;
 }
 catch ( ... ) {
  PyErr_SetString( coral::PyCoral::Exception(),
                  ( std::string( "An error occured when loading " ) + sinput ).c_str() );
  return 0;
 }
 Py_RETURN_NONE;
}


PyObject*
Context_setVerbosityLevel(  PyObject* self, PyObject* args )
{
 coral::PyCoral::Context* py_this = (coral::PyCoral::Context*) self;
 if( ! py_this->object ){
  PyErr_SetString( coral::PyCoral::Exception(),
                   (char*) "Error: No Valid C++ Object" );
  return 0;
 }
 try {
  const char* input = 0;
  if (PyString_Check(args))
   input = PyString_AsString(args);
  else{
   PyErr_SetString( coral::PyCoral::Exception(),
                   (char*) "Error: Bad Argument Type" );
   return 0;
  }
  const std::string sinput = std::string(input);
  std::vector< seal::Handle<seal::IMessageService> > v_msgSvc;
  py_this->object->query( v_msgSvc );
  if ( v_msgSvc.empty() ) {
   seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( py_this->object );
   if ( ! loader->load( "SEAL/Services/MessageService" ) ) {
    PyErr_SetString( coral::PyCoral::Exception(),
                     (char*) "Could not load the Message Service!" );
    return 0;
   }
  }
  py_this->object->query( v_msgSvc );
  if ( v_msgSvc.empty() ) {
   PyErr_SetString( coral::PyCoral::Exception(),
                   (char*) "Could not load the Message Service!" );
   return 0;
  }
  seal::Handle<seal::IMessageService>& msgSvc = v_msgSvc.front();
  if ( sinput == "VERBOSE" || sinput == "Verbose" || sinput == "verbose" || sinput == "V" || sinput == "v" ) msgSvc->setOutputLevel( seal::Msg::Verbose );
  else if ( sinput == "DEBUG" || sinput == "Debug" || sinput == "debug" || sinput == "D" || sinput == "d" ) msgSvc->setOutputLevel( seal::Msg::Debug );
  else if ( sinput == "INFO" || sinput == "Info" || sinput == "info" || sinput == "I" || sinput == "i" ) msgSvc->setOutputLevel( seal::Msg::Info );
  else if ( sinput == "WARNING" || sinput == "Warning" || sinput == "warning" || sinput == "W" || sinput == "w" ) msgSvc->setOutputLevel( seal::Msg::Warning );
  else if ( sinput == "ERROR"  || sinput == "Error" || sinput == "error" || sinput == "E" || sinput == "e" ) msgSvc->setOutputLevel( seal::Msg::Error );
  else if ( sinput == "FATAL" || sinput == "Fatal" || sinput == "fatal" || sinput == "F" || sinput == "f" ) msgSvc->setOutputLevel( seal::Msg::Fatal );
  else if ( sinput == "ALWAYS" || sinput == "Always" || sinput == "always" || sinput == "A" || sinput == "a" ) msgSvc->setOutputLevel( seal::Msg::Always );
  else
   msgSvc->setOutputLevel( seal::Msg::Nil );
 }
 catch ( std::exception& e ) {
  PyErr_SetString( coral::PyCoral::Exception(), e.what() );
  return 0;
 }
 catch ( ... ) {
  PyErr_SetString( coral::PyCoral::Exception(),
                  (char*) "An error occured when setting the verbosity level" );
  return 0;
 }
 Py_RETURN_NONE;
}
