#include "Connection.h"
#include "Session.h"
#include "ConnectionProperties.h"
#include "DomainProperties.h"
#include "CoralCommon/MonitoringEventDescription.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/SessionException.h"
#include "SealKernel/Service.h"
#include "SealKernel/Context.h"
#include "SealKernel/MessageStream.h"
#include "SealBase/SearchPath.h"
#include "sqlite3.h"
coral::SQLiteAccess::Connection::Connection( const coral::SQLiteAccess::DomainProperties& domainProperties, const std::string& connectionString):
  m_properties( new coral::SQLiteAccess::ConnectionProperties( domainProperties, connectionString ) ),m_isConnected(false),m_serverVersion( SQLITE_VERSION ){}

coral::SQLiteAccess::Connection::~Connection()
{
  if ( this->isConnected() )
    this->disconnect();
  delete m_properties;
}

coral::ISession*
coral::SQLiteAccess::Connection::newSession( const std::string& schemaName,
                                             coral::AccessMode mode ) const
{
  if ( ! const_cast<coral::SQLiteAccess::Connection*>( this )->isConnected() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(), "IConnection::newSession" );
  //m_properties->setReadOnly( ( mode == coral::Update ) ? false : true );
  //if ( m_properties->isReadOnly() && !m_inFileName.isReadable() ){
  if ( mode==coral::ReadOnly && !m_properties->databaseFileName().exists() ){
    const_cast<coral::SQLiteAccess::Connection*>(this)->disconnect();
    throw coral::DatabaseNotAccessibleException(std::string(" cannot open read-only session to non-existing ")+m_properties->databaseFileName().name(),"Connection::newSession",m_properties->domainProperties().service()->name() );   
  } 
  if( mode==coral::Update && !m_properties->databaseFileName().directory().exists() ){
    const_cast<coral::SQLiteAccess::Connection*>(this)->disconnect();
    throw coral::DatabaseNotAccessibleException(std::string("directory ")+std::string(m_properties->databaseFileName().directory().name())+std::string(" does not exist"), "Connection::newSession",m_properties->domainProperties().service()->name());
  }
  // Check if the access mode is compatible
  /*if ( mode == coral::Update && m_properties->isReadOnly() )
    throw coral::InvalidOperationInReadOnlyModeException( m_properties->domainProperties().service()->name(), "IConnection::newSession" );
  */
  if(this->size()>0){
    //const_cast<coral::SQLiteAccess::Connection*>(this)->disconnect();
    throw coral::MaximumNumberOfSessionsException(m_properties->domainProperties().service()->name());
  }
  return new coral::SQLiteAccess::Session( *( static_cast<coral::IDevConnection*>( const_cast<coral::SQLiteAccess::Connection*>( this ) ) ), *m_properties,schemaName,mode );
}

void
coral::SQLiteAccess::Connection::connect()
{
  if ( this->isConnected() ) {
    seal::MessageStream log( m_properties->domainProperties().service(),
                             m_properties->domainProperties().service()->name() );
    log << seal::Msg::Warning << "A connection is already present for \""
        << m_properties->connectionString() << "\"" << seal::flush;
  }
  /**
     throw coral::DatabaseNotAccessibleException if the file does not exist and the directory is unwritable
  */
  const std::string database = m_properties->connectionString();
  this->verifyFileName(database);
  if( !m_properties->databaseFileName().exists() && !m_properties->databaseFileName().directory().isWritable() ){
    throw coral::DatabaseNotAccessibleException(m_properties->databaseFileName().directory().name()+std::string(" is not writable"),"Connection::connect",m_properties->domainProperties().service()->name() ); 
  }
  m_isConnected=true; //set logical connection

  //Record the beginning of the connection using the Monitoring service
  std::vector< seal::IHandle< coral::monitor::IMonitoringService > > loadedMonitoringServices;
  m_properties->domainProperties().service()->context()->query( loadedMonitoringServices );
  if ( ! loadedMonitoringServices.empty() ) {
    coral::monitor::IMonitoringService& monitoringService = *( loadedMonitoringServices.front().get() );
    monitoringService.record( m_properties->connectionString(),
			      coral::monitor::Session,
			      coral::monitor::Info,
			      monitoringEventDescription.connectionBegin() );
  }
  return;
}

bool
coral::SQLiteAccess::Connection::isConnected( bool /* probePhysicalConnection */)
{
  return m_isConnected;
}


void
coral::SQLiteAccess::Connection::disconnect()
{
  if ( this->isConnected() ) {
    // Clean up all the active sessions !!!
    this->invalidateAll(); ;
    // Record the ending of the connection
  }
  m_isConnected=false;
  // Record the ending of the connection
  std::vector< seal::IHandle< coral::monitor::IMonitoringService > > loadedMonitoringServices;
  m_properties->domainProperties().service()->context()->query( loadedMonitoringServices );
  if ( ! loadedMonitoringServices.empty() ) {
    coral::monitor::IMonitoringService& monitoringService = *( loadedMonitoringServices.front().get() );
    monitoringService.record( m_properties->connectionString(),
			      coral::monitor::Session,
			      coral::monitor::Info,
			      monitoringEventDescription.connectionEnd() );
  }
}

std::string
coral::SQLiteAccess::Connection::serverVersion() const
{
  if ( ! const_cast<coral::SQLiteAccess::Connection*>( this )->isConnected() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(),
                                               "IConnection::serverVersion" );
  return m_serverVersion;
}


coral::ITypeConverter&
coral::SQLiteAccess::Connection::typeConverter()
{
  if ( ! this->isConnected() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(),
                                               "Connection::typeConverter" );
  return m_properties->typeConverter();
}

void 
coral::SQLiteAccess::Connection::verifyFileName( const std::string& inFileName )const{
  const seal::Filename fileNameFirstGuess( inFileName );
  const char* thePathVariable = ::getenv( "SQLITE_FILE_PATH" );
  if( fileNameFirstGuess.isAbsolute(true) ){
    //m_inFileName=fileNameFirstGuess;
    m_properties->setFileName(fileNameFirstGuess);
    return;
  }else if( fileNameFirstGuess.isRelative() ){
    m_properties->setFileName(fileNameFirstGuess.expand(seal::Filename(".")));
    //m_inFileName=fileNameFirstGuess.expand(seal::Filename("."));
    return;
  }
  seal::SearchPath path( thePathVariable );
  for ( seal::SearchPath::const_iterator iDirectory = path.begin();
	      iDirectory != path.end(); ++iDirectory ) {
    const seal::Filename guess( *iDirectory, fileNameFirstGuess );
    m_properties->setFileName(guess);
    //m_inFileName=guess;
    return;
  }
  m_properties->setFileName(fileNameFirstGuess);
  //m_inFileName=fileNameFirstGuess;
  return;
}
