#include "SQLiteStatement.h"
#include "DomainProperties.h"
#include "SessionProperties.h"
#include "SQLiteCommon.h"
#include "StatementStatistics.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/ITypeConverter.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Date.h"
#include "CoralBase/TimeStamp.h"
#include "CoralBase/Exception.h"
#include "SealBase/TimeInfo.h"
#include "SealKernel/MessageStream.h"
#include "SealKernel/Service.h"
#include "sqlite3.h"
#include <cstdlib>
#include <algorithm>
#include <climits>


namespace coral{
  SQLiteAccess::SQLiteStatement::SQLiteStatement(const SessionProperties& properties) :
    m_properties(properties),
    m_stmt(0),
    m_statementStatistics(0),
    m_rollBack( false )
  {
    if (m_properties.monitoringService()) m_statementStatistics = new coral::SQLiteAccess::StatementStatistics(); 
  }
  bool SQLiteAccess::SQLiteStatement::prepare( const std::string& statement ){
    if ( statement == "ROLLBACK" ) {
      m_rollBack = true;
    }
    //#ifdef _DEBUG
    //else {
    seal::MessageStream log( m_properties.domainProperties().service(),
                             m_properties.domainProperties().service()->name(),
                             seal::Msg::Debug );  
    log<<seal::Msg::Debug<<"SQLiteStatement::prepare "<<statement<<seal::flush;
    
    //#endif
    if(m_statementStatistics) m_statementStatistics->sqlStatement=statement;
    int rs=sqlite3_prepare( m_properties.dbHandle(), statement.c_str(), statement.length(), &m_stmt,0);
    if( rs!=SQLITE_OK ){
      //  seal::MessageStream log( m_properties.domainProperties().service(),
      //m_properties.domainProperties().service()->name() );
      log<<seal::Msg::Error<<"SQLiteStatement::prepare "<<rs<<" "<<sqlite3_errmsg( m_properties.dbHandle() )<<seal::flush;
      return false;
    }
    return true;
  }
  int SQLiteAccess::SQLiteStatement::step(){
    //m_log<<seal::Msg::Debug<<"SQLiteStatement::step"<<seal::flush;
    int rs=sqlite3_step(m_stmt);
    return rs;
  }
  bool SQLiteAccess::SQLiteStatement::bind( const coral::AttributeList& inputData ){
    seal::MessageStream log( m_properties.domainProperties().service(),
                             m_properties.domainProperties().service()->name(),
                             seal::Msg::Debug );  
    if((int)inputData.size() != sqlite3_bind_parameter_count(m_stmt) ){      
      log<<seal::Msg::Error<<"SQLiteStatement::bind wrong number of bind variables"<<seal::flush;
      return false;
    }
    log<<seal::Msg::Debug<<"SQLiteStatement::bind variables ";
    int rs=0;
    for(coral::AttributeList::iterator iAttribute=inputData.begin();
        iAttribute!=inputData.end();++iAttribute){
      std::string attributeName = iAttribute->specification().name();
      //std::cout<<attributeName<<std::endl;
      int idx=sqlite3_bind_parameter_index(m_stmt, (std::string(":")+attributeName).c_str());
      if(iAttribute!=inputData.begin()){
        log<<seal::Msg::Debug<<" , ";
      }
      log<<seal::Msg::Debug<<attributeName;
      //      std::cout<<"attributeName: "<<idx<<std::endl;
      //checking nulless
      if( iAttribute->isNull() ){
        rs=sqlite3_bind_null(m_stmt,idx);
        log<<seal::Msg::Debug<<" NULL ";
        continue;
      }
      const std::type_info& attributeType = iAttribute->specification().type(); 
      SQLiteAccess::SQLiteType st;
      void* addr=iAttribute->addressOfData();
      if(st(attributeType)==SQLT_INT){
        SQLiteAccess::castAttributeToInt c;
        int value=c(addr,attributeType);
        rs=sqlite3_bind_int(m_stmt,idx,value );
        log<<seal::Msg::Debug<<"(int) "<<value<<" ";
      }else if(st(attributeType)==SQLT_UIN){
        SQLiteAccess::castAttributeToUInt c;
        long long value=c(addr,attributeType);
        rs=sqlite3_bind_int64(m_stmt,idx,value );
        log<<seal::Msg::Debug<<"(unsigned int) "<<value<<" ";
      }else if(st(attributeType)==SQLT_64INT){
        SQLiteAccess::castAttributeTo64Int c;
        long long int value=c(addr,attributeType);
        rs=sqlite3_bind_int64(m_stmt,idx,value);
        log<<seal::Msg::Debug<<"(long long int) "<<value<<" ";
      }else if(st(attributeType)==SQLT_64UIN){
        SQLiteAccess::castAttributeTo64UInt c;
        unsigned long long int value=c(addr,attributeType);
        rs=sqlite3_bind_int64(m_stmt,idx,value);
        log<<seal::Msg::Debug<<"(unsigned long long int) "<<value<<" ";
      }else if(st(attributeType)==SQLT_DOUBLE){
        SQLiteAccess::castAttributeToDouble c;
        double value=c(addr,attributeType);
        rs=sqlite3_bind_double(m_stmt,idx,value);
        log<<seal::Msg::Debug<<"(double) "<<value<<" ";
      }else if(st(attributeType)==SQLT_TEXT){
        std::string val=(*iAttribute).data<std::string>();
        rs=sqlite3_bind_text(m_stmt,idx,val.c_str(),val.length(),SQLITE_TRANSIENT);
        log<<seal::Msg::Debug<<"(string) "<<val<<" ";
      }else if(st(attributeType)==SQLT_BLOB){
        long blobsize=iAttribute->data<coral::Blob>().size();
	if( blobsize==0 ){
	  const char key='a';
	  //rs=sqlite3_bind_zeroblob(m_stmt,idx, 0);
	  rs=sqlite3_bind_blob(m_stmt,idx,&key,0,SQLITE_TRANSIENT);
	}else{
	  void* blobdataaddress=iAttribute->data<coral::Blob>().startingAddress();
	  rs=sqlite3_bind_blob(m_stmt, idx,blobdataaddress,(int)blobsize,SQLITE_TRANSIENT);
	}
        log<<seal::Msg::Debug<<"(BLOB) of size "<<blobsize<<" ";
      }else if(st(attributeType)==SQLT_DATE ){
        seal::Time::ValueType value=(*iAttribute).data<coral::Date>().time().ns();
        rs=sqlite3_bind_int64(m_stmt,idx,(long long int)value);
        log<<seal::Msg::Debug<<"(DATE) "<<value<<" ";
      }else if(st(attributeType)==SQLT_TIMESTAMP){
        seal::Time::ValueType value=(*iAttribute).data<coral::TimeStamp>().time().ns();
        rs=sqlite3_bind_int64(m_stmt,idx,(long long int)value);
        log<<seal::Msg::Debug<<"(TIME) "<<value<<" ";
      }
      if(rs!=SQLITE_OK ){
        //seal::MessageStream log( m_properties.domainProperties().service(),
        //                         m_properties.domainProperties().service()->name() );
        log<<seal::Msg::Error<<"SQLiteStatement::bind "<<rs<<" "<<sqlite3_errmsg( m_properties.dbHandle() )<<seal::flush;
        return false;
      }
    }
    log<<seal::Msg::Debug<<seal::flush;
    return true;
  }
  bool SQLiteAccess::SQLiteStatement::setNumberOfPrefetchedRows( unsigned int /* numberOfRows */ ){
    //dummy now
    return true;
  }
  bool SQLiteAccess::SQLiteStatement::setCacheSize( unsigned int /* sizeInMB */ ){
    //dummy now
    return true;
  }
  bool SQLiteAccess::SQLiteStatement::execute(){
    //m_log<<seal::Msg::Debug<<"SQLiteStatement::execute"<<seal::flush;
    seal::TimeInfo::NanoSecs t0=0, t1=0;
    if ( m_statementStatistics ) t0 = seal::TimeInfo::processIdleTime();
    int rs=this->step();
    if( rs==SQLITE_DONE ){
      m_rowcounter=(unsigned int)sqlite3_changes( m_properties.dbHandle() );
      if ( m_statementStatistics ) {
        t1 = seal::TimeInfo::processIdleTime();
        double idleTimeSeconds = ( t1-t0 ) * 1e-9;
        m_statementStatistics->idleTime = idleTimeSeconds;
        m_properties.monitoringService()->record( m_properties.connectionString(), coral::monitor::Statement,coral::monitor::Time,m_statementStatistics->sqlStatement, m_statementStatistics->idleTime);
      }
      return true;
    }
    if ( m_statementStatistics ) {
      t1 = seal::TimeInfo::processIdleTime();
      double idleTimeSeconds = ( t1-t0 ) * 1e-9;
      m_statementStatistics->idleTime = idleTimeSeconds;
      m_properties.monitoringService()->record( m_properties.connectionString(), coral::monitor::Statement,coral::monitor::Time,m_statementStatistics->sqlStatement, m_statementStatistics->idleTime);
    }
    if ( ! m_rollBack ) {
      // seal::MessageStream log( m_properties.domainProperties().service(),
      // m_properties.domainProperties().service()->name() );  
      //log<<seal::Msg::Error<<"SQLiteStatement::execute "<<rs<<" "<<sqlite3_errmsg(  m_properties.dbHandle() )<<seal::flush;
      sqlite3_finalize(m_stmt);
      throw coral::Exception(sqlite3_errmsg(  m_properties.dbHandle() ),"SQLiteStatement::execute",m_properties.domainProperties().service()->name());
    }
    return false;
  }
  
  bool SQLiteAccess::SQLiteStatement::fetchNext(){
    seal::TimeInfo::NanoSecs t0=0;
    if ( m_statementStatistics ) t0 = seal::TimeInfo::processIdleTime();
    int rs=this->step();
    if( rs==SQLITE_ROW ){
      if ( m_statementStatistics ) {
        seal::TimeInfo::NanoSecs t1 = seal::TimeInfo::processIdleTime();
        double idleTimeSeconds = ( t1-t0 ) * 1e-9;
        m_statementStatistics->idleTime += idleTimeSeconds;
      }
      m_rowcounter++;
      return true;
    }
    if( rs==SQLITE_DONE ){
      if ( m_statementStatistics ) { 
        m_properties.monitoringService()->record( m_properties.connectionString(),
                                                  coral::monitor::Statement,
                                                  coral::monitor::Time,m_statementStatistics->sqlStatement,
                                                  m_statementStatistics->idleTime );
      }
      this->reset();
      return false;
    }
    seal::MessageStream log( m_properties.domainProperties().service(),
                             m_properties.domainProperties().service()->name() );  
    log<<seal::Msg::Error<<"SQLiteStatement::fetchNext "<<rs<<" "<<sqlite3_errmsg( m_properties.dbHandle() )<<seal::flush;///should throw here!
    if( m_statementStatistics ){
      m_properties.monitoringService()->record( m_properties.connectionString(), coral::monitor::Statement, coral::monitor::Time,m_statementStatistics->sqlStatement, m_statementStatistics->idleTime );
    }
    return false;
  }
  
  bool SQLiteAccess::SQLiteStatement::defineOutput( coral::AttributeList& outputData ){
    //m_log<<seal::Msg::Debug<<"SQLiteStatement::defineOutput"<<seal::flush;
    unsigned int idx = 0;
    if ( outputData.size() == 0 ) return true;
    for ( coral::AttributeList::iterator iAttribute = outputData.begin();
          iAttribute != outputData.end(); ++iAttribute, ++idx ) {
      if(sqlite3_column_type(m_stmt,idx)==SQLITE_NULL){
        iAttribute->setNull(true);
      }else{
        iAttribute->setNull(false);
        const std::type_info& attributeType = iAttribute->specification().type();
        SQLiteAccess::SQLiteType st;
        if(st(attributeType)==SQLT_INT){
          int result=0;
          result=sqlite3_column_int(m_stmt,idx);
          if(attributeType==typeid(int)){
            iAttribute->data<int>()=result;
          }else if( attributeType==typeid(short)){
            iAttribute->data<short>()=(short)result;
          }else if( attributeType==typeid(char)){
            iAttribute->data<char>()=(char)result;
          } else if( attributeType==typeid(long)){
            iAttribute->data<long>()=(long)result;
          }
        }else if(st(attributeType)==SQLT_UIN){
          unsigned int result=0;
          result=(unsigned int)sqlite3_column_int(m_stmt,idx);
          if( attributeType==typeid(unsigned short)){
            iAttribute->data<unsigned short>()=(unsigned short)result;
          }else if( attributeType==typeid(bool)){
            iAttribute->data<bool>()=(bool)result;
          }else if( attributeType==typeid(unsigned char)) {
            iAttribute->data<unsigned char>()=(unsigned char)result;
          }
        }else if(st(attributeType)==SQLT_64INT){
          long long int result=0;
          result=sqlite3_column_int64(m_stmt,idx);
          if(attributeType==typeid(long long int)){
            iAttribute->data<long long int>()=result;
          }else if( attributeType==typeid(long) ){
            iAttribute->data<long>()=(long)result;
          }
        }else if(st(attributeType)==SQLT_64UIN){
          unsigned long long int result=0;
          result=(unsigned long long int)sqlite3_column_int64(m_stmt,idx);
          if(attributeType==typeid(unsigned int)){
            iAttribute->data<unsigned int>()=result;
          }else if( attributeType==typeid(unsigned long)){
            iAttribute->data<unsigned long>()=(unsigned long)result;
          }else if(attributeType==typeid(unsigned long long int)) {
            iAttribute->data<unsigned long long int>()=result;
          }
        }else if(st(attributeType)==SQLT_DOUBLE){
          double result=0.0;
          result=sqlite3_column_double(m_stmt,idx);
          if(attributeType==typeid(double)){
            iAttribute->data<double>()=result;
          }else if( attributeType==typeid(float)){
            iAttribute->data<float>()=(float)result;
          }else if( attributeType==typeid(long double)){
            iAttribute->data<long double>()=(long double)result;
          }
        }else if(st(attributeType)==SQLT_TEXT){
          const char* result="";
          result=(const char*)sqlite3_column_text(m_stmt,idx);
          iAttribute->data<std::string>()=std::string(result);
        }else if(st(attributeType)==SQLT_BLOB){
          const void* result=sqlite3_column_blob(m_stmt,idx);
          int s=sqlite3_column_bytes(m_stmt,idx);
          coral::Blob& attrblob=iAttribute->data<coral::Blob>();
          attrblob.resize(s);
          void* attrblobad=attrblob.startingAddress();
          ::memcpy(attrblobad,result,s);
        }else if(st(attributeType)==SQLT_DATE || st(attributeType)==SQLT_TIMESTAMP){
          long long int result=0;
          result=sqlite3_column_int64(m_stmt,idx);
          seal::Time t((seal::Time::ValueType)result);
          if( st(attributeType)==SQLT_DATE ){
            iAttribute->data<coral::Date>()=coral::Date(t);
          }else{
            iAttribute->data<coral::TimeStamp>()=coral::TimeStamp(t);
          }
        }else{
          seal::MessageStream log( m_properties.domainProperties().service(),
                                   m_properties.domainProperties().service()->name() );  
          log << seal::Msg::Error << "Unsupported type : \"" + iAttribute->specification().typeName() + "\"" << seal::flush;
          this->reset();
          return false;
        }
      }//end-of-if
    }//end-of-for
    return true;
  }
  unsigned int SQLiteAccess::SQLiteStatement::numberOfRowsProcessed() const{
    //m_log<<seal::Msg::Debug<<"SQLiteStatement::numberOfRowsProcessed "<<m_rowcounter<<seal::flush;
    return m_rowcounter;
  }

  const std::type_info*  SQLiteAccess::SQLiteStatement::decltypeForColumn( int columnId ) const{
    /*#ifdef _DEBUG
      seal::MessageStream log( m_properties.domainProperties().service(),
      m_properties.domainProperties().service()->name(),
      seal::Msg::Nil );  
      log<<seal::Msg::Debug<<"SQLiteStatement::decltypeForColumn "<<columnId<<seal::flush;
      #endif
    */
    //const coral::ITypeConverter& typeConverter=m_properties.typeConverter();
    const char* t=sqlite3_column_decltype(m_stmt, columnId);
    return coral::AttributeSpecification::typeIdForName( m_properties.typeConverter().cppTypeForSqlType(std::string(t)) ); 
  }
  
  bool SQLiteAccess::SQLiteStatement::reset(){
    //m_log<<seal::Msg::Debug<<"SQLiteStatement::reset"<<seal::flush;
    int rs=sqlite3_reset(m_stmt);
    if(rs!=SQLITE_OK ){
      seal::MessageStream log( m_properties.domainProperties().service(),
                               m_properties.domainProperties().service()->name() );  
      log<<seal::Msg::Error<<"SQLiteStatement::reset "<<rs<<" "<<sqlite3_errmsg(  m_properties.dbHandle() )<<seal::flush;
      return false;
    }
    m_rowcounter=0;
    if(m_statementStatistics) m_statementStatistics->reset();
    return true;
  } 
  
  SQLiteAccess::SQLiteStatement::~SQLiteStatement(){
    //m_log<<seal::Msg::Debug<<"SQLiteStatement::~SQLiteStatement"<<seal::flush;
    int rs=sqlite3_finalize(m_stmt);
    if(rs!=SQLITE_OK ){
      if ( ! m_rollBack ) {
        seal::MessageStream log( m_properties.domainProperties().service(),
                                 m_properties.domainProperties().service()->name() );  
        log<<seal::Msg::Error<<"SQLiteStatement::~SQLiteStatement "<<rs<<" "<<sqlite3_errmsg(  m_properties.dbHandle() )<<seal::flush;
      }
    }
    if(m_statementStatistics) delete m_statementStatistics;
  }
}//ns coral
