#include "Session.h"
#include "SessionProperties.h"
#include "ConnectionProperties.h"
#include "DomainProperties.h"
#include "MonitorController.h"
#include "Transaction.h"
#include "Schema.h"
#include "CoralCommon/URIParser.h"
#include "CoralCommon/MonitoringEventDescription.h"
#include "RelationalAccess/MonitoringException.h"
#include "RelationalAccess/SessionException.h"
#include "RelationalAccess/IAuthenticationService.h"
#include "RelationalAccess/IAuthenticationCredentials.h"
#include "RelationalAccess/IMonitoringService.h"
#include "SealKernel/Service.h"
#include "SealKernel/Context.h"
#include "SealKernel/MessageStream.h"
#include "SealBase/SearchPath.h"
//#include "SealBase/FileSystemError.h"
#include "sqlite3.h"
//#include <iostream>
coral::SQLiteAccess::Session::Session(coral::IDevConnection& connection,
				      coral::SQLiteAccess::ConnectionProperties& connectionProperties,
                                      const std::string& /* schemaName */, coral::AccessMode mode):
  coral::IDevSession( connection ),
  m_properties( new coral::SQLiteAccess::SessionProperties(connectionProperties,static_cast<coral::ISession&>( *this ),( mode == coral::ReadOnly ) ? true : false ) ) ,
  m_monitorController( new coral::SQLiteAccess::MonitorController( *m_properties ) ),
  m_transaction(0),
  m_schema(0),
  m_dbHandle(0)
{}
coral::SQLiteAccess::Session::~Session()
{
  if ( this->isUserSessionActive() ) this->endUserSession();  
  delete m_monitorController;
  delete m_properties;
}
coral::IMonitoring&
coral::SQLiteAccess::Session::monitoring()
{
  return *m_monitorController;
}
void
coral::SQLiteAccess::Session::startUserSession( const std::string& /* userName */ ,
						const std::string&  /* password */){
  if ( this->isUserSessionActive() ) return;
  if ( m_properties->isReadOnly() && !m_properties->connectionProperties().databaseFileName().isReadable() ){
    throw coral::AuthenticationFailureException( m_properties->domainProperties().service()->name() );
  }
  if( sqlite3_open(m_properties->connectionProperties().databaseFileName().name(), &m_dbHandle) ){
    this->endUserSession();
    throw coral::SessionException( "Could not open file "+std::string(m_properties->connectionProperties().databaseFileName().name()), "Session::startUserSession", m_properties->domainProperties().service()->name() );
  }
  m_properties->setdbHandle(m_dbHandle);
  // Create new schema and transaction objects.
  m_schema = new coral::SQLiteAccess::Schema( *m_properties );
  m_transaction = new coral::SQLiteAccess::Transaction( *m_properties );
  // Record the beginning of the session
  if ( m_properties->monitoringService() ) {
    try{
      m_properties->monitoringService()->record( m_properties->connectionString(),
                                                 coral::monitor::Session,
                                                 coral::monitor::Info,
                                                 monitoringEventDescription.sessionBegin() );
    }catch(const coral::MonitoringException& e){
      seal::MessageStream log( m_properties->domainProperties().service(),
                               m_properties->domainProperties().service()->name() );  
      log<<seal::Msg::Error<<"caught  coral::MonitoringException in Session::startUserSession"<<e.what()<<seal::flush;
    }
  }
  return;
}

void
coral::SQLiteAccess::Session::endUserSession()
{
  if ( this->isUserSessionActive() ) {
    // Abort any active transaction
    if ( this->transaction().isActive() ) this->transaction().rollback();
    if ( m_transaction ) {
      delete m_transaction;
      m_transaction = 0;
    }
    if ( m_schema ) {
      delete m_schema;
      m_schema = 0;
    }
    if ( m_properties->monitoringService() ) {
      m_properties->monitoringService()->record( m_properties->connectionString(),
						 coral::monitor::Session,
						 coral::monitor::Info,
						 monitoringEventDescription.sessionEnd() );
    }
    sqlite3_close( m_dbHandle );
    m_dbHandle=0;
    m_properties->setdbHandle(0);
  }
}


bool
coral::SQLiteAccess::Session::isUserSessionActive() const
{
  return ( m_dbHandle!=0 );
}

coral::ITransaction&
coral::SQLiteAccess::Session::transaction()
{
  if ( ! this->isUserSessionActive() ) {
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name(), "ISession::transaction" );
  }
  return *m_transaction;
}
coral::ISchema& 
coral::SQLiteAccess::Session::nominalSchema(){
  if ( ! this->isUserSessionActive() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name() );
  return *m_schema;
}
coral::ISchema&
coral::SQLiteAccess::Session::schema( const std::string& /* schemaName */)
{
  if ( ! this->isUserSessionActive() )
    throw coral::ConnectionNotActiveException( m_properties->domainProperties().service()->name() );
  return *m_schema;
}

sqlite3* 
coral::SQLiteAccess::Session::getdbHandle() const
{
  return m_dbHandle;
}
