#include "InsertSelect.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/IConnection.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IOperationWithQuery.h"
#include "RelationalAccess/IBulkOperationWithQuery.h"
#include "RelationalAccess/IQueryDefinition.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/SchemaException.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralCommon/Utilities.h"
#include <stdexcept>
#include <memory>
#include <iostream>


InsertSelect::InsertSelect( coral::IRelationalDomain& domain,
                            const std::string& connectionString,
                            coral::IAuthenticationService& authenticationService ):
  TestBase( authenticationService, connectionString ),
  m_connection( 0 ),
  m_schemaName( "" )
{
  std::pair<std::string,std::string> databaseAndSchema = domain.decodeUserConnectionString( connectionString );
  m_connection = domain.newConnection( databaseAndSchema.first );
  m_schemaName = databaseAndSchema.second;
}


InsertSelect::~InsertSelect()
{
  if ( m_connection ) delete m_connection;
}


bool
InsertSelect::run()
{
  try {
    m_connection->connect();
    std::auto_ptr<coral::ISession> session( m_connection->newSession( m_schemaName ) );
    session->startUserSession( this->username(), this->password() );
    session->transaction().start();

    coral::ISchema& workingSchema = session->nominalSchema();
    workingSchema.dropIfExistsTable( "T4" );
    coral::TableDescription description( workingSchema.tableHandle( "T2" ).description() );
    description.setName( "T4" );
    coral::ITable& table = workingSchema.createTable( description );
    table.privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );

    coral::IOperationWithQuery* operation = table.dataEditor().insertWithQuery();
    coral::IQueryDefinition& query = operation->query();
    query.addToTableList( "T2" );
    query.setCondition( "id<10", coral::AttributeList() );
    long numberOfRowsInserted = operation->execute();
    delete operation;
    if ( numberOfRowsInserted != 6 )
      throw std::runtime_error( "Unexpected number of rows inserted." );

    coral::IBulkOperationWithQuery* bulkOperation = table.dataEditor().bulkInsertWithQuery( 3 );
    coral::IQueryDefinition& queryb = bulkOperation->query();
    queryb.addToTableList( "T2" );
    coral::AttributeList condition;
    condition.extend<long>( "idmin");
    condition.extend<long>( "idmax");
    queryb.setCondition( "id>:idmin AND id<:idmax", condition );
    long& idmin = condition[0].data<long>();
    long& idmax = condition[1].data<long>();

    idmin = 10;
    idmax = 13;
    for ( int i = 0; i < 5; ++i ) {
      bulkOperation->processNextIteration();
      idmin += 4;
      idmax += 4;
    }
    bulkOperation->flush();
    delete bulkOperation;

    // add three more rows (for the next tests to make better sense)
    coral::AttributeList rowBuffer;

    table.dataEditor().rowBuffer( rowBuffer );
    rowBuffer[0].data<long long>() = 100;
    rowBuffer[1].data<float>() = 200;
    rowBuffer[2].data<double>() = 300;
    for ( int i = 0; i < 3; ++i ) {
      table.dataEditor().insertRow( rowBuffer );
      rowBuffer[0].data<long long>() += 1;
      rowBuffer[1].data<float>() += 2;
      rowBuffer[2].data<double>() += 3;
    }

    session->transaction().commit();
    session->endUserSession();
    m_connection->disconnect();
    coral::sleepSeconds( 1 );
  }
  catch ( coral::SchemaException& e ) {
    std::cerr << "Schema exception : " << e.what() << std::endl;
    return false;
  }
  return true;
}
