#include "SetOperations.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/IConnection.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/SchemaException.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"

#include <stdexcept>
#include <memory>
#include <iostream>


SetOperations::SetOperations( coral::IRelationalDomain& domain,
                              const std::string& connectionString,
                              coral::IAuthenticationService& authenticationService,
                              const std::string& workingSchema ):
  TestBase( authenticationService, connectionString ),
  m_connection( 0 ),
  m_schemaName( "" ),
  m_workingSchema( workingSchema )
{
  std::pair<std::string,std::string> databaseAndSchema = domain.decodeUserConnectionString( connectionString );
  m_connection = domain.newConnection( databaseAndSchema.first );
  m_schemaName = databaseAndSchema.second;
}


SetOperations::~SetOperations()
{
  if ( m_connection ) delete m_connection;
}


bool
SetOperations::run()
{
  try {
    m_connection->connect();
    std::auto_ptr<coral::ISession> session( m_connection->newSession( m_schemaName ) );
    session->startUserSession( this->username(), this->password() );
    session->transaction().start( true );

    coral::ISchema& workingSchema = session->schema( m_workingSchema );

    std::cout << " Query 1:" << std::endl;
    coral::IQuery* query1 = workingSchema.newQuery();
    query1->addToTableList( "T2" );
    coral::IQueryDefinition& rhs1 = query1->applySetOperation( coral::IQueryDefinition::Union );
    rhs1.addToTableList( "T4" );
    query1->setRowCacheSize( 100 );
    int nRows = 0;
    coral::ICursor& cursor1 = query1->execute();
    while ( cursor1.next() ) {
      cursor1.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query1;
    if ( nRows != 28 )
      throw std::runtime_error( "Unexpected number of rows" );

    std::cout << " Query 2:" << std::endl;
    coral::IQuery* query2 = workingSchema.newQuery();
    query2->addToTableList( "T4" );
    coral::IQueryDefinition& rhs2 = query2->applySetOperation( coral::IQueryDefinition::Minus );
    rhs2.addToTableList( "T2" );
    nRows = 0;
    coral::ICursor& cursor2 = query2->execute();
    while ( cursor2.next() ) {
      cursor2.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query2;
    if ( nRows != 3 )
      throw std::runtime_error( "Unexpected number of rows" );


    std::cout << " Query 3:" << std::endl;
    coral::IQuery* query3 = workingSchema.newQuery();
    query3->addToTableList( "T2" );
    coral::IQueryDefinition& rhs3 = query3->applySetOperation( coral::IQueryDefinition::Minus );
    rhs3.addToTableList( "T4" );
    nRows = 0;
    coral::ICursor& cursor3 = query3->execute();
    while ( cursor3.next() ) {
      cursor3.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query3;
    if ( nRows != 12 )
      throw std::runtime_error( "Unexpected number of rows" );

    std::cout << " Query 4:" << std::endl;
    coral::IQuery* query4 = workingSchema.newQuery();
    query4->addToTableList( "T2" );
    coral::IQueryDefinition& rhs4 = query4->applySetOperation( coral::IQueryDefinition::Intersect );
    rhs4.addToTableList( "T4" );
    nRows = 0;
    coral::ICursor& cursor4 = query4->execute();
    while ( cursor4.next() ) {
      cursor4.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query4;
    if ( nRows != 13 )
      throw std::runtime_error( "Unexpected number of rows" );


    std::cout << " Query 5:" << std::endl;
    coral::IQuery* query5 = workingSchema.newQuery();

    coral::IQueryDefinition& sub = query5->defineSubQuery( "A" );
    sub.addToTableList( "T2" );
    sub.setCondition( "id>10", coral::AttributeList() );
    coral::IQueryDefinition& rhs5 = sub.applySetOperation( coral::IQueryDefinition::Intersect );
    rhs5.addToTableList( "T4" );

    query5->addToTableList( "A", "B" );
    query5->addToOrderList( "ty-tx" );
    query5->limitReturnedRows( 4 );
    nRows = 0;
    coral::ICursor& cursor5 = query5->execute();
    while ( cursor5.next() ) {
      cursor5.currentRow().toOutputStream( std::cout ) << std::endl;
      ++nRows;
    }
    delete query5;
    if ( nRows != 4 )
      throw std::runtime_error( "Unexpected number of rows" );

    session->transaction().commit();
  }
  catch ( coral::SchemaException& e ) {
    std::cerr << "Schema exception : " << e.what() << std::endl;
    return false;
  }
  return true;
}
