#include "TestApp.h"
#include "SchemaDefinition.h"
#include "DmlOperations.h"
#include "Queries.h"
#include "InsertSelect.h"
#include "SetOperations.h"
#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/IAuthenticationService.h"
#include "SealKernel/Context.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/IMessageService.h"
#include "PluginManager/PluginManager.h"
#include <stdexcept>


TestApp::TestApp( const std::string& writerConnectionString,
                  const std::string& readerConnectionString,
                  const std::string& readerWorkingSchema ):
  m_context( new seal::Context ),
  m_writerConnectionString( writerConnectionString ),
  m_readerConnectionString( readerConnectionString ),
  m_readerWorkingSchema( readerWorkingSchema )
{
  this->loadServices();
}


TestApp::~TestApp()
{
}


void
TestApp::loadServices()
{
  seal::PluginManager* pm = seal::PluginManager::get();
  pm->initialise();
  seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( m_context.get() );

  // Load the SEAL Message service and set the output threshold to Info
  loader->load( "SEAL/Services/MessageService" );
  std::vector< seal::Handle<seal::IMessageService> > v_msgSvc;
  m_context->query( v_msgSvc );
  if ( ! v_msgSvc.empty() ) {
    seal::Handle<seal::IMessageService>& msgSvc = v_msgSvc.front();
    msgSvc->setOutputLevel( seal::Msg::Info);
  }

  // Load the XML Authentication Service
  loader->load( "CORAL/Services/XMLAuthenticationService" );

  // Load the Relational Service
  loader->load( "CORAL/Services/RelationalService" );
}


coral::IRelationalDomain&
TestApp::domain( const std::string& connectionString )
{
  std::vector< seal::IHandle<coral::IRelationalService> > v_svc;
  m_context->query( v_svc );
  if ( v_svc.empty() ) {
    throw std::runtime_error( "Could not locate the relational service" );
  }
  seal::IHandle<coral::IRelationalService>& relationalService = v_svc.front();
  return relationalService->domainForConnection( connectionString );
}


void
TestApp::run()
{
  std::vector< seal::IHandle<coral::IAuthenticationService> > v_svc;
  m_context->query( v_svc );
  if ( v_svc.empty() ) {
    throw std::runtime_error( "Could not locate the authentication service" );
  }

  coral::IAuthenticationService& authenticationService = *( v_svc.front().get() );

  std::cout << "Test 1: defining and reading back tables and views" << std::endl;
  SchemaDefinition* test1 = new SchemaDefinition( this->domain( m_writerConnectionString ),
                                                  m_writerConnectionString,
                                                  authenticationService );
  bool result = test1->createSchema() && test1->readSchema();
  delete test1;
  if ( ! result )
    throw std::runtime_error( "Test failed" );

  std::cout << "Test 2: DML operations" << std::endl;
  DmlOperations* test2 = new DmlOperations( this->domain( m_writerConnectionString ),
                                            m_writerConnectionString,
                                            authenticationService );
  result = test2->atomicOperations() && test2->bulkOperations();
  delete test2;
  if ( ! result )
    throw std::runtime_error( "Test failed" );

  std::cout << "Test 3: Queries" << std::endl;
  Queries* test3 = new Queries( this->domain( m_readerConnectionString ),
                                m_readerConnectionString,
                                authenticationService,
                                m_readerWorkingSchema );
  result = test3->run();
  delete test3;
  if ( ! result )
    throw std::runtime_error( "Test failed" );

  std::cout << "Test 4: Insert/Select statements" << std::endl;
  InsertSelect* test4 = new InsertSelect( this->domain( m_writerConnectionString ),
                                          m_writerConnectionString,
                                          authenticationService );
  result = test4->run();
  delete test4;
  if ( ! result )
    throw std::runtime_error( "Test failed" );

  std::cout << "Test 5: Queries involving set operations" << std::endl;
  SetOperations* test5 = new SetOperations( this->domain( m_writerConnectionString ),
                                            m_readerConnectionString,
                                            authenticationService,
                                            m_readerWorkingSchema );
  result = test5->run();
  delete test5;
  if ( ! result )
    throw std::runtime_error( "Test failed" );
}
