#include "TestApp.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/IMessageService.h"
#include "PluginManager/PluginManager.h"
#include <stdexcept>
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/IReplicaSortingAlgorithm.h"
#include "RelationalAccess/IDatabaseServiceDescription.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITransaction.h"
#include "CoralBase/Exception.h"
#include "CoralBase/AttributeSpecification.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/IConnectionService.h"
#include "RelationalAccess/IConnectionServiceConfiguration.h"
#include "RelationalAccess/ConnectionServiceException.h"
#include "CoralCommon/Utilities.h"

const std::string TestApp::OVALTAG = "[OVAL]";

class KeepOnlyOneReplica : public coral::IReplicaSortingAlgorithm {
  public:
    KeepOnlyOneReplica(unsigned int replicaIndex):m_replicaIndex(replicaIndex){}
    
    virtual ~KeepOnlyOneReplica(){}
    
    void sort(std::vector<const coral::IDatabaseServiceDescription*>& replicaSet){
      std::vector<const coral::IDatabaseServiceDescription*> vec;
      if(m_replicaIndex+1 <= replicaSet.size()){
        vec.push_back(replicaSet[m_replicaIndex]);
      }
      replicaSet = vec;
    }

    void setReplicaIndex(unsigned int replicaIndex){
      m_replicaIndex = replicaIndex;
    }


  private:
    unsigned int m_replicaIndex;
};

TestApp::TestApp():
  m_context( new seal::Context ){
  this->loadServices();
}


TestApp::~TestApp()
{
}

void
TestApp::loadServices()
{
  seal::PluginManager* pm = seal::PluginManager::get();
  pm->initialise();
  seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( m_context.get() );

  // Load the SEAL Message service and set the output threshold to Info
  loader->load( "SEAL/Services/MessageService" );
  std::vector< seal::Handle<seal::IMessageService> > v_msgSvc;
  m_context->query( v_msgSvc );
  if ( ! v_msgSvc.empty() ) {
    seal::Handle<seal::IMessageService>& msgSvc = v_msgSvc.front();
    msgSvc->setOutputLevel( seal::Msg::Debug);
  }

  // Load the Connection Service
  std::cout << "Loading Connection Service"<<std::endl;
  loader->load( "CORAL/Services/ConnectionService" );
}

seal::IHandle<coral::IConnectionService>
TestApp::connectionService()
{
  std::vector< seal::IHandle<coral::IConnectionService> > v_svc;
  m_context->query( v_svc );
  if ( v_svc.empty() ) {
    throw std::runtime_error( "Could not locate the connection service" );
  }
  return v_svc.front();
}

//
void TestApp::setupDataSource( const std::string& connectionString){
  seal::IHandle<coral::IConnectionService> theConnectionService = connectionService();
  coral::ISessionProxy* session = theConnectionService->connect( connectionString );
  session->transaction().start();
  // creates a dummy table to be looked up
  session->nominalSchema().dropIfExistsTable("T_CONNECTION_SERVICE_TEST");
  coral::TableDescription descr;
  descr.setName("T_CONNECTION_SERVICE_TEST");
  descr.insertColumn("N_X",coral::AttributeSpecification::typeNameForType<int>());
  descr.insertColumn("N_S",coral::AttributeSpecification::typeNameForType<std::string>());
  session->nominalSchema().createTable( descr );
  session->transaction().commit();
  delete session;
  coral::sleepSeconds(1);
}

void
TestApp::run(){
  
  seal::IHandle<coral::IConnectionService> theConnectionService = connectionService();
  if(! theConnectionService ){
    std::cout << "Connection Service not loaded."<<std::endl;
    return;
  }
  theConnectionService->configuration().setConnectionTimeOut(0);
  KeepOnlyOneReplica* sorter = new KeepOnlyOneReplica(0);
  theConnectionService->configuration().setReplicaSortingAlgorithm(sorter);
  
  setupDataSource("oracle://lcg_coral_nightly/coral_writer");

  theConnectionService->purgeConnectionPool();
  
  try {
    coral::ISessionProxy* session = 0;
    std::string connectionString = "/ReplicaHandling/dbdev/service_3";  
    session = theConnectionService->connect( connectionString, coral::ReadOnly  );
    session->transaction().start( true );
    std::set<std::string> tables = session->nominalSchema().listTables();
    session->transaction().commit();  
    delete session;
    session = 0;
    sorter->setReplicaIndex( 1 );
    connectionString = "/ReplicaHandling/dbdev/service_3";  
    session = theConnectionService->connect( connectionString, coral::Update  );
    session->transaction().start();
    tables = session->nominalSchema().listTables();
    if(tables.find("T_CONNECTION_SERVICE_TEST")==tables.end()){
      std::cout << OVALTAG <<"TEST ERROR with connection "<<connectionString<<" table T_CONNECTION_SERVICE_TEST is not found."<<std::endl;
    } else {
      std::cout << OVALTAG << "Connection "<<connectionString<<" was ok."<<std::endl;
    }
    session->transaction().commit();  
    delete session;
    session = 0;
    connectionString = "/ReplicaHandling/dbdev/service_4";  
    try {
      sorter->setReplicaIndex( 2 );
      session = theConnectionService->connect( connectionString, coral::Update  );
      session->transaction().start();
      tables = session->nominalSchema().listTables();
      session->transaction().commit();
      delete session;
    } catch (const coral::ConnectionServiceException& ex){
      std::cout << OVALTAG <<"Exception:"<<ex.what()<<std::endl;
    }  
  }catch ( const coral::Exception& exc) {
    std::cout << "TEST ERROR: "<<exc.what()<<std::endl;
  }
  
   
}
