#include "TestDriver.h"

#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IBulkOperation.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"

#include "CoralBase/Blob.h"
#include "CoralBase/TimeStamp.h"
#include "CoralBase/Date.h"

#include "SealKernel/Context.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/IMessageService.h"
#include "PluginManager/PluginManager.h"

#include <stdexcept>
#include <sstream>
#include <cfloat>
#include <cmath>


static const std::string tableName = "CORAL_TESTING_T_1";
static const int rowsWithSimpleInsert = 10;
static const int rowsWithBulkInsert = 55;


TestDriver::TestDriver():
  m_context( new seal::Context ),
  m_rowBuffer( 0 )
{
  seal::PluginManager::get()->initialise();
  seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( m_context.get() );
  loader->load("SEAL/Services/MessageService");
  std::vector< seal::Handle<seal::IMessageService> > v_msgSvc;
  m_context->query( v_msgSvc );
  if ( ! v_msgSvc.empty() ) {
    seal::Handle<seal::IMessageService>& msgSvc = v_msgSvc.front();
    msgSvc->setOutputLevel( seal::Msg::Warning);
  }
}


TestDriver::~TestDriver()
{
  if ( m_rowBuffer ) delete m_rowBuffer;
}


void
TestDriver::prepareSchema( const std::string& serviceName )
{
  coral::ConnectionService connectionService( m_context.get() );

  coral::ISessionProxy* session = connectionService.connect( serviceName, coral::Update );
  if ( ! session )
    throw std::runtime_error( "Could not connect to " + serviceName );

  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();

  // Clear the old table
  schema.dropIfExistsTable( tableName );

  if ( m_rowBuffer ) delete m_rowBuffer;
  m_rowBuffer = new coral::AttributeList;
  coral::AttributeList& rowBuffer = *m_rowBuffer;

  // Create the new table
  coral::TableDescription description;
  description.setName( tableName );

  description.insertColumn( "ID",
                            coral::AttributeSpecification::typeNameForId( typeid( int ) ) );
  description.setNotNullConstraint( "ID" );
  description.setPrimaryKey( "ID" );
  rowBuffer.extend<int>( "ID" );

  description.insertColumn( "V_B",
                            coral::AttributeSpecification::typeNameForId( typeid( bool ) ) );
  rowBuffer.extend<bool>( "V_B" );

  description.insertColumn( "V_UC",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned char ) ) );
  rowBuffer.extend<unsigned char>( "V_UC" );

  description.insertColumn( "V_SC",
                            coral::AttributeSpecification::typeNameForId( typeid( char ) ) );
  rowBuffer.extend<char>( "V_SC" );

  description.insertColumn( "V_US",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned short ) ) );
  rowBuffer.extend<unsigned short>( "V_US" );

  description.insertColumn( "V_SS",
                            coral::AttributeSpecification::typeNameForId( typeid( short ) ) );
  rowBuffer.extend<short>( "V_SS" );

  description.insertColumn( "V_UI",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned int ) ) );
  rowBuffer.extend<unsigned int>( "V_UI" );

  description.insertColumn( "V_SI",
                            coral::AttributeSpecification::typeNameForId( typeid( int ) ) );
  rowBuffer.extend<int>( "V_SI" );

  description.insertColumn( "V_UL",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned long ) ) );
  rowBuffer.extend<unsigned long>( "V_UL" );

  description.insertColumn( "V_SL",
                            coral::AttributeSpecification::typeNameForId( typeid( long ) ) );
  rowBuffer.extend<long>( "V_SL" );

  description.insertColumn( "V_ULL",
                            coral::AttributeSpecification::typeNameForId( typeid( unsigned long long ) ) );
  rowBuffer.extend<unsigned long long>( "V_ULL" );

  description.insertColumn( "V_SLL",
                            coral::AttributeSpecification::typeNameForId( typeid( long long ) ) );
  rowBuffer.extend<long long>( "V_SLL" );

  description.insertColumn( "V_F",
                            coral::AttributeSpecification::typeNameForId( typeid( float ) ) );
  rowBuffer.extend<float>( "V_F" );

  description.insertColumn( "V_D",
                            coral::AttributeSpecification::typeNameForId( typeid( double ) ) );
  rowBuffer.extend<double>( "V_D" );

  description.insertColumn( "V_LD",
                            coral::AttributeSpecification::typeNameForId( typeid( double ) ) );
  rowBuffer.extend<double>( "V_LD" );

  description.insertColumn( "V_BLOB",
                            coral::AttributeSpecification::typeNameForId( typeid( coral::Blob ) ) );
  rowBuffer.extend<coral::Blob>( "V_BLOB" );

  description.insertColumn( "V_DATE",
                            coral::AttributeSpecification::typeNameForId( typeid( coral::Date ) ) );
  rowBuffer.extend<coral::Date>( "V_DATE" );

  description.insertColumn( "V_TIMESTAMP",
                            coral::AttributeSpecification::typeNameForId( typeid( coral::TimeStamp ) ) );
  rowBuffer.extend<coral::TimeStamp>( "V_TIMESTAMP" );

  description.insertColumn( "V_SG",
                            coral::AttributeSpecification::typeNameForId( typeid( std::string ) ) );
  rowBuffer.extend<std::string>( "V_SG" );

  description.insertColumn( "V_SF",
                            coral::AttributeSpecification::typeNameForId( typeid( std::string ) ),
                            5 );
  rowBuffer.extend<std::string>( "V_SF" );

  description.insertColumn( "V_SV",
                            coral::AttributeSpecification::typeNameForId( typeid( std::string ) ),
                            50,
                            false );
  rowBuffer.extend<std::string>( "V_SV" );

  schema.createTable( description );

  session->transaction().commit();

  delete session;
}


void
TestDriver::writeSimple( const std::string& serviceName )
{
  coral::ConnectionService connectionService( m_context.get() );

  coral::ISessionProxy* session = connectionService.connect( serviceName, coral::Update );
  if ( ! session )
    throw std::runtime_error( "Could not connect to " + serviceName );

  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle( tableName );
  coral::ITableDataEditor& dataEditor = table.dataEditor();
  coral::AttributeList& rowBuffer = *m_rowBuffer;

  for ( int i = 0; i < rowsWithSimpleInsert; ++i ) {
    this->fillData( i );
    dataEditor.insertRow( rowBuffer );
  }

  session->transaction().commit();

  delete session;
}


void
TestDriver::writeBulk( const std::string& serviceName )
{
  coral::ConnectionService connectionService( m_context.get() );

  coral::ISessionProxy* session = connectionService.connect( serviceName, coral::Update );
  if ( ! session )
    throw std::runtime_error( "Could not connect to " + serviceName );

  session->transaction().start();

  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle( tableName );
  coral::ITableDataEditor& dataEditor = table.dataEditor();
  coral::AttributeList& rowBuffer = *m_rowBuffer;

  coral::IBulkOperation* bulkOperation = dataEditor.bulkInsert( rowBuffer,
                                                                rowsWithSimpleInsert );

  for ( int i = rowsWithSimpleInsert; i < rowsWithSimpleInsert + rowsWithBulkInsert; ++i ) {
    this->fillData( i );
    bulkOperation->processNextIteration();
  }
  bulkOperation->flush();

  delete bulkOperation;

  session->transaction().commit();

  delete session;
}


void
TestDriver::read( const std::string& serviceName )
{
  coral::ConnectionService connectionService( m_context.get() );

  coral::ISessionProxy* session = connectionService.connect( serviceName, coral::ReadOnly );
  if ( ! session )
    throw std::runtime_error( "Could not connect to " + serviceName );

  session->transaction().start( true );

  coral::ISchema& schema = session->nominalSchema();
  coral::ITable& table = schema.tableHandle( tableName );
  coral::IQuery* query = table.newQuery();
  query->addToOrderList( "ID" );
  query->defineOutput( *m_rowBuffer );
  query->setRowCacheSize( rowsWithSimpleInsert );

  coral::ICursor& cursor = query->execute();
  for ( int i = 0; i < rowsWithSimpleInsert + rowsWithBulkInsert; ++i ) {
    if ( ! cursor.next() )
      throw std::runtime_error( "Unexpected number of returned rows" );
    this->checkData( i );
  }
  if ( cursor.next() )
    throw std::runtime_error( "Unexpected number of returned rows" );

  delete query;

  session->transaction().commit();

  delete session;
}



void
TestDriver::fillData( int i )
{
  coral::AttributeList& row = *m_rowBuffer;

  row[0].data<int>() = i;

  if ( i%3 == 0 ) {
    row[1].setNull( true );
  }
  else {
    row[1].setNull( false );
    row[1].data<bool>() = ( i%2 == 0 ) ? true : false;
  }

  row[2].data<unsigned char>() = ( i*2 ) % 256;

  row[3].data<char>() = ( i*2 ) % 128;

  row[4].data<unsigned short>() = ( i*3 ) % 2048;

  row[5].data<short>() = ( i*3 ) % 2048 - 1024;

  row[6].data<unsigned int>() = 2147483648U - i;//( i*4 ) % 1000;

  row[7].data<int>() = ( i*4 ) % 1000 - 500;

  row[8].data<unsigned long>() = ( i*4 ) % 1001;

  row[9].data<long>() = ( i*4 ) % 1001 - 500;

  row[10].data<unsigned long long>() =  2147483648U - i;

  row[11].data<long long>() = ( (i%2 == 0 ) ? ( 2147483648ll - 2*i ) : ( i - 2147483648ll ) );

  if ( i%4 == 0 ) {
    row[12].setNull( true );
  }
  else {
    row[12].setNull( false );
    row[12].data<float>() = i + 0.123; 
  }

  row[13].data<double>() = 0.123456789 - 2.3*i; 

  row[14].data<double>() = 0.123456789 + 2.3*i; 

  coral::Blob& blob = row[15].data<coral::Blob>();
  int blobSize = 1000 * ( i%100 + 1 );
  blob.resize( blobSize );
  unsigned char* p = static_cast<unsigned char*>( blob.startingAddress() );
  for ( int j = 0; j < blobSize; ++j, ++p ) *p = ( i + j )%256;

  row[16].data<coral::Date>() = coral::Date( 2006, 1, i%31 + 1 );

  row[17].data<coral::TimeStamp>() = coral::TimeStamp( 2006, 1, 12, 15, 47, i%60, 0 );

  std::ostringstream os1;
  os1 << "A general String : " << i % 5;
  std::string s1 = os1.str();
  row[18].data<std::string>() = s1;

  std::ostringstream os2;
  os2 << "...." << i % 10;
  std::string s2 = os2.str();
  row[19].data<std::string>() = s2;

  std::ostringstream os3;
  os3 << "Value : " << i;
  std::string s3 = os3.str();
  row[20].data<std::string>() = s3;
}


void
TestDriver::checkData( int i ) const
{
  const coral::AttributeList& row = *m_rowBuffer;

  if ( row[0].data<int>() != i )
    throw std::runtime_error( "Unexpected value for variable " + row[0].specification().name() );

  if ( i%3 == 0 ) {
    if ( ! row[1].isNull() )
      throw std::runtime_error( "Unexpected NOT NULL data for variable " + row[1].specification().name() );
  }
  else {
    if ( row[1].isNull() )
      throw std::runtime_error( "Unexpected NULL data for variable " + row[1].specification().name() );
    if ( row[1].data<bool>() != ( ( i%2 == 0 ) ? true : false ) )
      throw std::runtime_error( "Unexpected value for variable " + row[1].specification().name() );
  }

  if ( row[2].data<unsigned char>() != ( i*2 ) % 256 )
    throw std::runtime_error( "Unexpected value for variable " + row[2].specification().name() );

  if ( row[3].data<char>() != ( i*2 ) % 128 )
    throw std::runtime_error( "Unexpected value for variable " + row[3].specification().name() );

  if ( row[4].data<unsigned short>() != ( i*3 ) % 2048 )
    throw std::runtime_error( "Unexpected value for variable " + row[4].specification().name() );

  if ( row[5].data<short>() != ( i*3 ) % 2048 - 1024 )
    throw std::runtime_error( "Unexpected value for variable " + row[5].specification().name() );

  if ( row[6].data<unsigned int>() != static_cast<unsigned int>( 2147483648U - i ) ) // ( i*4 ) % 1000 ) )
    throw std::runtime_error( "Unexpected value for variable " + row[6].specification().name() );

  if ( row[7].data<int>() != ( i*4 ) % 1000 - 500 )
    throw std::runtime_error( "Unexpected value for variable " + row[7].specification().name() );

  if ( row[8].data<unsigned long>() != static_cast<unsigned long>( ( i*4 ) % 1001 ) )
    throw std::runtime_error( "Unexpected value for variable " + row[8].specification().name() );

  if ( row[9].data<long>() != ( i*4 ) % 1001 - 500 )
    throw std::runtime_error( "Unexpected value for variable " + row[9].specification().name() );

  if ( row[10].data<unsigned long long>() != static_cast<unsigned long long>( 2147483648U - i ) )
    throw std::runtime_error( "Unexpected value for variable " + row[10].specification().name() );

  if ( row[11].data<long long>() != ( (i%2 == 0 ) ? ( 2147483648ll - 2*i ) : ( i - 2147483648ll ) ) )
    throw std::runtime_error( "Unexpected value for variable " + row[11].specification().name() );

  if ( i%4 == 0 ) {
    if ( ! row[12].isNull() )
      throw std::runtime_error( "Unexpected NOT NULL data for variable " + row[12].specification().name() );
  }
  else {
    if ( row[12].isNull() )
      throw std::runtime_error( "Unexpected NULL data for variable " + row[12].specification().name() );
    if ( std::abs( row[12].data<float>() / ( i + 0.123 ) - 1 ) > std::numeric_limits<float>::epsilon() )
      throw std::runtime_error( "Unexpected value for variable " + row[12].specification().name() );
  }

  if ( std::abs( row[13].data<double>() / ( 0.123456789 - 2.3*i ) - 1 ) > std::numeric_limits<double>::epsilon() )
    throw std::runtime_error( "Unexpected value for variable " + row[13].specification().name() );

  if ( std::abs( row[14].data<double>() / ( 0.123456789 + 2.3*i ) - 1 ) > std::numeric_limits<double>::epsilon() )
    throw std::runtime_error( "Unexpected value for variable " + row[14].specification().name() );

  const coral::Blob& blob = row[15].data<coral::Blob>();
  int blobSize = 1000 * ( i%100 + 1 );
  if ( blob.size() != blobSize )
    throw std::runtime_error( "Unexpected blob size for variable " + row[15].specification().name() );
  const unsigned char* p = static_cast<const unsigned char*>( blob.startingAddress() );
  for ( int j = 0; j < blobSize; ++j, ++p )
    if ( *p != ( i + j )%256 )
      throw std::runtime_error( "Unexpected value for variable " + row[15].specification().name() );

  if ( row[16].data<coral::Date>() != coral::Date( 2006, 1, i%31 + 1 ) )
    throw std::runtime_error( "Unexpected value for variable " + row[16].specification().name() );
  
  if ( row[17].data<coral::TimeStamp>() != coral::TimeStamp( 2006, 1, 12, 15, 47, i%60, 0 ) )
    throw std::runtime_error( "Unexpected value for variable " + row[17].specification().name() );

  std::ostringstream os1;
  os1 << "A general String : " << i % 5;
  std::string s1 = os1.str();
  if ( row[18].data<std::string>() != s1 )
    throw std::runtime_error( "Unexpected value for variable " + row[16].specification().name() );

  std::ostringstream os2;
  os2 << "...." << i % 10;
  std::string s2 = os2.str();
  if ( row[19].data<std::string>() != s2 )
    throw std::runtime_error( "Unexpected value for variable " + row[17].specification().name() );

  std::ostringstream os3;
  os3 << "Value : " << i;
  std::string s3 = os3.str();
  if ( row[20].data<std::string>() != s3 )
    throw std::runtime_error( "Unexpected value for variable " + row[18].specification().name() );
}
