#include "TestApp.h"

#include "RelationalAccess/IAuthenticationService.h"
#include "RelationalAccess/IAuthenticationCredentials.h"
#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/ISession.h"
#include "RelationalAccess/IConnection.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/IMonitoringReporter.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableSchemaEditor.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "SealKernel/Context.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/IMessageService.h"
#include "PluginManager/PluginManager.h"
#include "CoralCommon/Utilities.h"
#include <stdexcept>


TestApp::TestApp( const std::string& connectionString ):
  m_context( new seal::Context ),
  m_connectionString( connectionString )
{
  this->loadServices();
}


TestApp::~TestApp()
{
}


void
TestApp::loadServices()
{
  seal::PluginManager* pm = seal::PluginManager::get();
  pm->initialise();
  seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( m_context.get() );

  // Load the SEAL Message service and set the output threshold to Info
  loader->load( "SEAL/Services/MessageService" );
  std::vector< seal::Handle<seal::IMessageService> > v_msgSvc;
  m_context->query( v_msgSvc );
  if ( ! v_msgSvc.empty() ) {
    seal::Handle<seal::IMessageService>& msgSvc = v_msgSvc.front();
    msgSvc->setOutputLevel( seal::Msg::Info);
  }

  // Load the XML Authentication Service
  loader->load( "CORAL/Services/XMLAuthenticationService" );

  // Load the Monitoring service
  loader->load( "CORAL/Services/MonitoringService" );

  // Load the Relational Service
  loader->load( "CORAL/Services/RelationalService" );
}


coral::IRelationalDomain&
TestApp::domain( const std::string& connectionString )
{
  std::vector< seal::IHandle<coral::IRelationalService> > v_svc;
  m_context->query( v_svc );
  if ( v_svc.empty() ) {
    throw std::runtime_error( "Could not locate the relational service" );
  }
  seal::IHandle<coral::IRelationalService>& relationalService = v_svc.front();
  return relationalService->domainForConnection( connectionString );
}


void
TestApp::run()
{
  try
  {
    this->writeData();  
    this->readData();
    this->report();
  }
  catch( const std::exception& e )
  {
    std::cerr << "Std. exception caught: " << e.what() << std::endl;
    throw;
  }
}

void TestApp::writeData()
{  
  seal::IHandle<coral::IAuthenticationService> authsvc = m_context->component( "CORAL/Services/XMLAuthenticationService" );

  if( ! authsvc )
    throw std::runtime_error( "Could not locate authentication service!" );

  const coral::IAuthenticationCredentials& crds = authsvc->credentials( m_connectionString );
  
  std::pair<std::string,std::string> cs = this->domain( m_connectionString ).decodeUserConnectionString( m_connectionString );

  std::auto_ptr<coral::IConnection> connection( this->domain( m_connectionString ).newConnection( cs.first ) );

  connection->connect();
  
  std::auto_ptr<coral::ISession> session( connection->newSession( cs.second, coral::Update ) );  

  session->monitoring().start( coral::monitor::Trace );

  session->startUserSession( crds.valueForItem("user"), crds.valueForItem("password") );

  session->transaction().start();

  // Create the table
  session->nominalSchema().dropIfExistsTable( "T10" );
  coral::TableDescription description1( "SchemaDefinition_Test" );
  description1.setName( "T10" );
  description1.insertColumn( "id",
                             coral::AttributeSpecification::typeNameForId( typeid(long) ) );
  // Define primary key
  description1.setPrimaryKey( "id" );
  description1.insertColumn( "t",
                             coral::AttributeSpecification::typeNameForId( typeid(short) ) );
  description1.insertColumn( "Xx",
                             coral::AttributeSpecification::typeNameForId( typeid(float) ) );
  // Set column not null
  description1.setNotNullConstraint( "Xx" );
  description1.insertColumn( "Y",
                             coral::AttributeSpecification::typeNameForId( typeid(double) ) );
  // Add a unique constraint specifying a name
  description1.setUniqueConstraint( "Y", "U_T10_Y" );
  description1.insertColumn( "Z",
                             coral::AttributeSpecification::typeNameForId( typeid(double) ) );
  // Add a variable size string
  description1.insertColumn( "Comment",
                             coral::AttributeSpecification::typeNameForId( typeid(std::string) ),
                             100,
                             false );
  // Add a blob
  description1.insertColumn( "Data",
                             coral::AttributeSpecification::typeNameForId( typeid(coral::Blob) ) );

  // Create the actual table
  coral::ITable& table = session->nominalSchema().createTable( description1 );


  // Change the column name
  table.schemaEditor().renameColumn( "Xx", "X" );

  // Change the column type
  table.schemaEditor().changeColumnType( "t",
                                         coral::AttributeSpecification::typeNameForId( typeid(long long) ) );
  // Add a not null constaint
  table.schemaEditor().setNotNullConstraint( "t" );

  // Add a unique constraint made out of two columns
  std::vector< std::string > constraintColumns( 2 );
  constraintColumns[0] = "X";
  constraintColumns[1] = "Z";
  table.schemaEditor().setUniqueConstraint( constraintColumns );

  /// Add a unique index
  table.schemaEditor().createIndex( "T10_IDX_t", "t", true );

  // Give privileges
  table.privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );

  coral::ITableDataEditor& editor1 = table.dataEditor();
  coral::AttributeList rowBuffer1;

  rowBuffer1.extend<long>( "id" );
  long& id = rowBuffer1[ "id" ].data<long>();

  rowBuffer1.extend< long long >( "t" );
  long long& t = rowBuffer1[ "t" ].data<long long>();

  rowBuffer1.extend< float >( "X" );
  float& x = rowBuffer1[ "X" ].data<float>();

  rowBuffer1.extend< double >( "Y" );
  double& y = rowBuffer1[ "Y" ].data<double>();

  rowBuffer1.extend< double >( "Z" );
  double& z = rowBuffer1[ "Z" ].data<double>();

  rowBuffer1.extend< std::string >( "Comment" );
  std::string& comment = rowBuffer1[ "Comment" ].data<std::string>();

  rowBuffer1.extend< coral::Blob >( "Data" );
  coral::Blob& blob = rowBuffer1[ "Data" ].data<coral::Blob>();

  for ( int i = 0; i < 10; ++i ) {
    id = i + 1;
    t = 1;
    t <<= 4*i;
    x = 1 + (i+1)*0.001;
    y = 2 + (i+1)*0.0001;
    z = i*1.1;
    std::ostringstream os;
    os << "Row " << i + 1;
    comment = os.str();

    int blobSize = 1000 * ( i + 1 );
    blob.resize( blobSize );
    unsigned char* p = static_cast<unsigned char*>( blob.startingAddress() );
    for ( int j = 0; j < blobSize; ++j, ++p ) *p = ( i + j )%256;

    if ( i%4 == 2 ) {
      rowBuffer1[ "Z" ].setNull( true );
    }
    else {
      rowBuffer1[ "Z" ].setNull( false );
    }

    // Insert the row
    editor1.insertRow( rowBuffer1 );
  }

  session->transaction().commit();

  coral::sleepSeconds( 1 );
}


void
TestApp::readData()
{
  seal::IHandle<coral::IAuthenticationService> authsvc = m_context->component( "CORAL/Services/XMLAuthenticationService" );

  if( ! authsvc )
    throw std::runtime_error( "Could not locate authentication service!" );

  const coral::IAuthenticationCredentials& crds = authsvc->credentials( m_connectionString );
  
  std::pair<std::string,std::string> cs = this->domain( m_connectionString ).decodeUserConnectionString( m_connectionString );

  std::auto_ptr<coral::IConnection> connection( this->domain( m_connectionString ).newConnection( cs.first ) );
  
  connection->connect();

  std::auto_ptr<coral::ISession> session( connection->newSession( cs.second, coral::ReadOnly ) );
  
  session->monitoring().start( coral::monitor::Trace );

  session->startUserSession( crds.valueForItem("user"), crds.valueForItem("password") );

  session->transaction().start( true );

  for ( int i = 0; i < 10; ++i ) {
    coral::IQuery* query0 = session->nominalSchema().tableHandle( "T10" ).newQuery();
    query0->addToOutputList( "Data" );
    query0->addToOutputList( "Z" );
    query0->addToOrderList( "id" );
    coral::ICursor& cursor0 = query0->execute();
    while ( cursor0.next() );
    if ( i == 5 )
      session->monitoring().stop();
    delete query0;
  }

  session->transaction().commit();
}


void
TestApp::report()
{
  std::pair<std::string,std::string> cs = this->domain( m_connectionString ).decodeUserConnectionString( m_connectionString );

  std::vector< seal::IHandle<coral::monitor::IMonitoringService> > v_svc;
  m_context->query( v_svc );
  if ( v_svc.empty() ) {
    throw std::runtime_error( "Could not locate the monitoring service" );
  }
  seal::IHandle<coral::monitor::IMonitoringService>& monitoringService = v_svc.front();

  std::cout << "Monitoring report for connection" << std::endl;
  std::string::size_type pos = m_connectionString.rfind( '/' );
  std::string connstr = m_connectionString.substr( 0, pos );
  monitoringService->reporter().report( connstr );

  std::cout << "Monitoring report for session" << std::endl;
  monitoringService->reporter().report( m_connectionString );
}
