#include "XMLAuthenticationService.h"
#include "DataSourceEntry.h"
#include "CoralCommon/AuthenticationCredentials.h"
#include "RelationalAccess/AuthenticationServiceException.h"
#include "SealKernel/PropertyManager.h"
#include "SealKernel/MessageStream.h"
#include "SealBase/SearchPath.h"
#include "SealBase/Filename.h"
#include "xercesc/parsers/XercesDOMParser.hpp"
#include "xercesc/dom/DOM.hpp"
#include "xercesc/sax/HandlerBase.hpp"
#include "xercesc/util/XMLString.hpp"
#include "xercesc/util/PlatformUtils.hpp"
#include <memory>
#include <cstdlib>

namespace coral {

  namespace XMLAuthenticationService {

    class CallbackForInputFileName {
      public:
        explicit CallbackForInputFileName( coral::XMLAuthenticationService::XMLAuthenticationService& authenticationService )
          : m_authenticationService( authenticationService )
        {
        }

        void operator() ( const seal::PropertyBase& p )
        {
          std::string fileName = boost::any_cast< std::string >( p.get() );
          m_authenticationService.setInputFileName( fileName );
        }
      private:
        coral::XMLAuthenticationService::XMLAuthenticationService& m_authenticationService;
    };
  }  
}

DEFINE_SEAL_COMPONENT ( coral::XMLAuthenticationService::XMLAuthenticationService, "CORAL/Services/XMLAuthenticationService")


coral::XMLAuthenticationService::XMLAuthenticationService::XMLAuthenticationService( seal::Context* c )
  : seal::Service( c, classContextKey() ),
    m_isInitialized( false ),
    m_inputFileName( "" ),
    m_data()
{
  propertyManager()->declareProperty( "AuthenticationFile",
                                      m_inputFileName,
                                      std::string("authentication.xml"),
                                      "The input file with the authentication data",
                                      coral::XMLAuthenticationService::CallbackForInputFileName( *this ) );
}


coral::XMLAuthenticationService::XMLAuthenticationService::XMLAuthenticationService( seal::Context* c,
                                                                                     const std::string& key )
  : seal::Service( c, key ),
    m_isInitialized( false ),
    m_inputFileName( "" ),
    m_data()
{
  propertyManager()->declareProperty( "AuthenticationFile",
                                      m_inputFileName,
                                      std::string("authentication.xml"),
                                      "The input file with the authentication data",
                                      coral::XMLAuthenticationService::CallbackForInputFileName( *this ) );
}


coral::XMLAuthenticationService::XMLAuthenticationService::~XMLAuthenticationService()
{
  for ( std::map< std::string, coral::XMLAuthenticationService::DataSourceEntry* >::iterator iConnection = m_data.begin();
        iConnection != m_data.end(); ++iConnection ) delete iConnection->second;
}


void
coral::XMLAuthenticationService::XMLAuthenticationService::setInputFileName(  const std::string& inputFileName )
{
  m_inputFileName = inputFileName;
  for ( std::map< std::string, coral::XMLAuthenticationService::DataSourceEntry* >::iterator iConnection = m_data.begin();
        iConnection != m_data.end(); ++iConnection ) delete iConnection->second;
  m_data.clear();
  m_isInitialized = false;
}


bool
coral::XMLAuthenticationService::XMLAuthenticationService::processFile( const std::string& inputFileName )
{
  bool result = true;
  try
  {
    xercesc::XercesDOMParser parser;
    parser.setValidationScheme( xercesc::XercesDOMParser::Val_Always );
    parser.setDoNamespaces( true );
    
    xercesc::HandlerBase errorHandler; 
    parser.setErrorHandler( &errorHandler );

    parser.parse( inputFileName.c_str() );

    xercesc::DOMDocument* document = parser.getDocument();

    XMLCh tempStr[20];
    xercesc::XMLString::transcode( "connection", tempStr, 19);

    xercesc::DOMNodeList* connectionList = document->getElementsByTagName( tempStr );
    
    if ( connectionList )
    {
      XMLSize_t numberOfConnections = connectionList->getLength();
      
      for ( XMLSize_t iConnection = 0; iConnection < numberOfConnections; ++iConnection )
      {
        xercesc::DOMNode* connectionNode = connectionList->item( iConnection );

        if ( connectionNode )
        {
          char*       connectionName  = xercesc::XMLString::transcode( connectionNode->getAttributes()->item( 0 )->getNodeValue() );
          std::string sConnectionName = connectionName;
          xercesc::XMLString::release( &connectionName );

          // Locate the credential
          coral::XMLAuthenticationService::DataSourceEntry* credential = 0;
          std::map< std::string, coral::XMLAuthenticationService::DataSourceEntry* >::iterator iConnection = m_data.find( sConnectionName );
          if ( iConnection != m_data.end() ) {
            credential = iConnection->second;
            // Issue a warning here.
            seal::MessageStream log( this, this->name() );
            log << seal::Msg::Warning << "Credential parameters for connection string \""
                << sConnectionName
                << "\" have already been defined. Only new elements are appended, while existing will be ignored."
                << seal::flush;
          }
          else {
            credential = new coral::XMLAuthenticationService::DataSourceEntry( this->name(), sConnectionName );
            m_data.insert( std::make_pair( sConnectionName, credential ) );
          }

          xercesc::DOMNodeList* parameterList = connectionNode->getChildNodes();
          
          if ( parameterList )
          {
            XMLSize_t numberOfParameters = parameterList->getLength();
            
            for ( XMLSize_t iParameter = 0; iParameter < numberOfParameters; ++iParameter )
            {
              xercesc::DOMNode* parameterNode = parameterList->item( iParameter );

              if ( parameterNode && parameterNode->getNodeType() == xercesc::DOMNode::ELEMENT_NODE )
              {
                char* nodeName = xercesc::XMLString::transcode( parameterNode->getNodeName() );
                std::string sNodeName = nodeName;
                xercesc::XMLString::release( &nodeName );
                
                if ( sNodeName == "parameter" ) { // The default parameters
                  char* parameterName = xercesc::XMLString::transcode( parameterNode->getAttributes()->item( 0 )->getNodeValue() );
                  std::string sParameterName = parameterName;
                  xercesc::XMLString::release( &parameterName );
                  char* parameterValue = xercesc::XMLString::transcode( parameterNode->getAttributes()->item( 1 )->getNodeValue() );
                  std::string sParameterValue = parameterValue;
                  xercesc::XMLString::release( &parameterValue );

                  credential->appendCredentialItem( sParameterName, sParameterValue );
                }
                else if ( sNodeName == "role" ) { // A role
                  char* roleName  = xercesc::XMLString::transcode( parameterNode->getAttributes()->item( 0 )->getNodeValue() );
                  std::string sRoleName = roleName;
                  xercesc::XMLString::release( &roleName );

                  // Retrieve the parameters for the role
                  xercesc::DOMNodeList* roleParameterList = parameterNode->getChildNodes();


                  if ( roleParameterList )
                  {
                    XMLSize_t numberOfRoleParameters = roleParameterList->getLength();
            
                    for ( XMLSize_t iRoleParameter = 0; iRoleParameter < numberOfRoleParameters; ++iRoleParameter )
                    {
                      xercesc::DOMNode* roleParameterNode = roleParameterList->item( iRoleParameter );
                      if ( roleParameterNode && roleParameterNode->getNodeType() == xercesc::DOMNode::ELEMENT_NODE )
                      {
                        char* roleNodeName = xercesc::XMLString::transcode( roleParameterNode->getNodeName() );
                        std::string sRoleNodeName = roleNodeName;
                        xercesc::XMLString::release( &roleNodeName );
                        
                        if ( sRoleNodeName == "parameter" ) {
                          char* roleParameterName = xercesc::XMLString::transcode( roleParameterNode->getAttributes()->item( 0 )->getNodeValue() );
                          std::string sRoleParameterName = roleParameterName;
                          xercesc::XMLString::release( &roleParameterName );
                          char* roleParameterValue = xercesc::XMLString::transcode( roleParameterNode->getAttributes()->item( 1 )->getNodeValue() );
                          std::string sRoleParameterValue = roleParameterValue;
                          xercesc::XMLString::release( &roleParameterValue );
                          
                          credential->appendCredentialItemForRole( sRoleParameterName, sRoleParameterValue, sRoleName );
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }

    parser.reset();
  }
  catch ( const xercesc::XMLException& toCatch )
  {
    char* message = xercesc::XMLString::transcode( toCatch.getMessage() );
    seal::MessageStream log( this, this->name() );
    log << seal::Msg::Error << message << seal::flush;
    xercesc::XMLString::release( &message );
    result = false;
  }
  catch ( const xercesc::DOMException& toCatch )
  {
    char* message = xercesc::XMLString::transcode( toCatch.msg );
    seal::MessageStream log( this, this->name() );
    log << seal::Msg::Error << message << seal::flush;
    xercesc::XMLString::release( &message );
    result = false;
  }
  catch (...)
  {
    seal::MessageStream log( this, this->name() );
    log << seal::Msg::Error << "Unexpected Exception parsing file \"" << inputFileName << "\"" << seal::flush;
    result = false;
  }

  return result;
}


bool
coral::XMLAuthenticationService::XMLAuthenticationService::initialize()
{
  std::vector< std::string > inputFileNames = this->verifyFileName();
  if ( inputFileNames.empty() )
  {
    seal::MessageStream log( this, this->name() );
    log << seal::Msg::Error << "Could not open \"" << m_inputFileName << "\" for reading" << seal::flush;
    return false;
  }

  try
  {
    xercesc::XMLPlatformUtils::Initialize();
  }
  catch ( const xercesc::XMLException& toCatch )
  {
    char* message = xercesc::XMLString::transcode( toCatch.getMessage() );
    seal::MessageStream log( this, this->name() );
    log << seal::Msg::Error << message << seal::flush;
    xercesc::XMLString::release( &message );
    return false;
  }

  bool result = true;
  for ( std::vector< std::string >::const_iterator iFileName = inputFileNames.begin();
        iFileName != inputFileNames.end(); ++iFileName ) {
    if ( ! this->processFile( *iFileName ) ) {
      result = false;
      break;
    }
  }

  xercesc::XMLPlatformUtils::Terminate();

  m_isInitialized = result;
  return result;
}


const coral::IAuthenticationCredentials&
coral::XMLAuthenticationService::XMLAuthenticationService::credentials( const std::string& connectionString ) const
{
  boost::mutex::scoped_lock lock(m_mutexLock);
  if ( ! m_isInitialized ) {
    const_cast< coral::XMLAuthenticationService::XMLAuthenticationService* >( this )->initialize();
  }
  std::map< std::string, coral::XMLAuthenticationService::DataSourceEntry* >::const_iterator iConnection = m_data.find( connectionString );
  if ( iConnection == m_data.end() )
    throw coral::UnknownConnectionException( this->name(), connectionString );
  return iConnection->second->credentials();
}


const coral::IAuthenticationCredentials&
coral::XMLAuthenticationService::XMLAuthenticationService::credentials( const std::string& connectionString,
                                                                        const std::string& role ) const
{
  boost::mutex::scoped_lock lock(m_mutexLock);
  if ( ! m_isInitialized ) {
    const_cast< coral::XMLAuthenticationService::XMLAuthenticationService* >( this )->initialize();
  }
  std::map< std::string, coral::XMLAuthenticationService::DataSourceEntry* >::const_iterator iConnection = m_data.find( connectionString );
  if ( iConnection == m_data.end() )
    throw coral::UnknownConnectionException( this->name(), connectionString );
  return iConnection->second->credentials( role );
}


std::vector< std::string >
coral::XMLAuthenticationService::XMLAuthenticationService::verifyFileName() const
{
  std::vector< std::string > fileNames;

  // Try the file name as is...
  seal::Filename fileNameFirstGuess( m_inputFileName );

  if ( fileNameFirstGuess.exists() ) {
    fileNames.push_back( m_inputFileName );
  }

  // If an absolute path has been specified stop there.
  if ( fileNameFirstGuess.isAbsolute() ) return fileNames;

  // Try to find other files in the path variable
  const char* thePathVariable = ::getenv( "CORAL_AUTH_PATH" );
  if ( ! thePathVariable ) return fileNames;

  seal::SearchPath path( thePathVariable );
  for ( seal::SearchPath::const_iterator iDirectory = path.begin();
        iDirectory != path.end(); ++iDirectory ) {
    seal::Filename guess( *iDirectory, fileNameFirstGuess );
    if ( guess.exists() ) {
      fileNames.push_back( guess.name() );
    }
  }
  return fileNames;
}
