#include <iostream>
#include <sstream>
#include <stdexcept>
#include "PluginManager/PluginManager.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/Context.h"
#include "SealKernel/Service.h"
#include "SealKernel/PropertyManager.h"
#include "RelationalAccess/ILookupService.h"
#include "RelationalAccess/IDatabaseServiceSet.h"
#include "RelationalAccess/IDatabaseServiceDescription.h"


void reportReplica( const coral::IDatabaseServiceDescription& replica,
                    const std::string& logicalName )
{
  std::cout << logicalName << " : "
            << replica.connectionString() << ", "
            << replica.authenticationMechanism() << ", "
            << ( ( replica.accessMode() == coral::ReadOnly ) ? std::string("Read-Only") : std::string("Update") )
            << std::endl;
  std::string connectionRetrialPeriodParam  = replica.serviceParameter("ConnectionRetrialPeriod");
  std::string connectionRetrialTimeOutParam = replica.serviceParameter("ConnectionRetrialTimeOut");
  std::string connectionTimeOutParam = replica.serviceParameter("ConnectionTimeOut");
  std::string missingConnectionExclusionTimeParam = replica.serviceParameter("MissingConnectionExclusionTime");
  if(!connectionRetrialPeriodParam.empty()) std::cout << "Par \"ConnectionRetrialPeriodParam\"="<<connectionRetrialPeriodParam<<std::endl;
  if(!connectionRetrialTimeOutParam.empty()) std::cout << "Par \"ConnectionRetrialTimeOutParam\"="<<connectionRetrialTimeOutParam<<std::endl;
  if(!connectionTimeOutParam.empty()) std::cout << "Par \"ConnectionTimeOutParam\"="<<connectionTimeOutParam<<std::endl;
  if(!missingConnectionExclusionTimeParam.empty()) std::cout << "Par \"MissingConnectionExclusionTimeParam\"="<<missingConnectionExclusionTimeParam<<std::endl;
  std::cout << "----------------------------------------"<<std::endl;
}



int main( int, char** )
{
  try {
    seal::PluginManager* pm = seal::PluginManager::get();
    pm->initialise();

    seal::Handle< seal::Context > context = new seal::Context();
    seal::Handle<seal::ComponentLoader> loader = new seal::ComponentLoader( context.get() );

    loader->load( "CORAL/Services/XMLLookupService" );

    seal::IHandle<seal::Service> svc = context->query<seal::Service>( "CORAL/Services/XMLLookupService" );

    seal::Handle<seal::PropertyManager>& pmgr = svc->propertyManager();
    seal::PropertyBase* property = pmgr->property( "LookupFile" );
    property->set( std::string( "dblookup.xml" ) );

    std::vector< seal::IHandle<coral::ILookupService> > v_lookupSvc;
    context->query( v_lookupSvc );

    if ( ! v_lookupSvc.empty() )
    {
      const coral::ILookupService& lookupSvc = *( v_lookupSvc.front() );


      coral::IDatabaseServiceSet* dbSet = lookupSvc.lookup( "myservice0" );
      int numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 0 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      delete dbSet;


      dbSet = lookupSvc.lookup( "myservice1" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 2 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice2" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 1 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice2" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice1", coral::ReadOnly );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 3 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice2", coral::ReadOnly );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 3 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice2" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice2", coral::ReadOnly, "password" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 2 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice2" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice3");
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 2 ) {
        std::cout << "Number of replica="<<numberOfReplicas<<std::endl;
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice3" );
      delete dbSet;

      std::cout << "Trying to load the dblookup.xml via HTTP..." << std::endl;
      
      // Try the HTTP based file access
      property->set( std::string( "http://rado.web.cern.ch/rado/dblookup.xml" ) );

      dbSet = lookupSvc.lookup( "myservice0" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 0 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice1" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 2 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice2" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 1 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice2" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice1", coral::ReadOnly );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 3 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice1" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice2", coral::ReadOnly );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 3 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice2" );
      delete dbSet;

      dbSet = lookupSvc.lookup( "myservice2", coral::ReadOnly, "password" );
      numberOfReplicas = dbSet->numberOfReplicas();
      if ( numberOfReplicas != 2 ) {
        delete dbSet;
        throw std::runtime_error( "Unexpected number of replicas retrieved" );
      }
      for ( int i = 0; i < numberOfReplicas; ++i )
        reportReplica( dbSet->replica( i ), "myservice2" );
      delete dbSet;
      
    }
    else {
      throw std::runtime_error("Could not retrieve an lookup service.");
    }

    std::cout << "Exiting..." << std::endl;
  }
  catch ( std::exception& e ) {
    std::cerr << "Standard C++ exception : " << e.what() << std::endl;
    return 1;
  }
  catch ( ... ) {
    std::cerr << "Exception caught (...)" << std::endl;
    return 1;
  }
  std::cout << "[OVAL] Success" << std::endl;
  return 0;
}
